#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import numpy


def softmax(sequence):
    """
    Compute the softmax of a sequence of values, i.e. the resulting probability distribution over that sequence.

    :param sequence: the sequence of which to compute the softmax
    :return: the probability distribution over the sequence
    """
    # Make sure the sequence is a numpy array
    if not isinstance(sequence, numpy.ndarray):
        sequence = numpy.array(sequence)
    # Make sure the length of the shape of the given array is 2
    assert len(sequence.shape) == 2
    # Get the element with max value in the given array as the normalization factor
    normalization_factor = numpy.max(sequence, axis=1)
    # Increase the shape size of the normalization factor to allow broadcasting
    normalization_factor = normalization_factor[:, numpy.newaxis]
    # Apply the normalization
    numerator = numpy.exp(sequence - normalization_factor)
    # Compute the denominator by summing all the normalized numerator elements
    denominator = numpy.sum(numerator, axis=1)
    # Increase the shape size of the denominator to allow broadcasting
    denominator = denominator[:, numpy.newaxis]
    # Return the softmax result probability distribution
    return numerator / denominator



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging

# Import required src

from usienarl import Environment, SpaceType


class Interface:
    """
    Base interface abstract class to define interaction between agent and environment.

    An interface allows to translate the environment state and actions to the agent observations and action.
    The default interface is the pass-through interface, implementing a simple fully observable interface for the
    environment. The environment is specific for each interface, and should not be changed outside of initialization.

    To define a specific interface, the abstract class should be implemented in a specific child class.
    """

    def __init__(self,
                 environment: Environment):
        # Define interface attribute
        self._environment: Environment = environment

    def get_random_agent_action(self,
                                logger: logging.Logger,
                                session):
        """
        Get a random action as seen by the agent, i.e. a random action in the environment translated in agent action.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :return: the random action as seen by the agent
        """
        # Get the random environment action
        environment_action = self._environment.get_random_action(logger, session)
        # Translate it to agent action and return it
        return self.environment_action_to_agent_action(logger, session, environment_action)

    def agent_action_to_environment_action(self,
                                           logger: logging.Logger,
                                           session,
                                           agent_action):
        """
        Translate the given agent action to the respective environment action.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param agent_action: the action as seen by the agent
        :return: the action as seen by the environment relative to the given agent action
        """
        raise NotImplementedError()

    def environment_action_to_agent_action(self,
                                           logger: logging.Logger,
                                           session,
                                           environment_action):
        """
        Translate the given environment action to the respective agent action.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param environment_action: the action as seen by the environment
        :return: the action as seen by the agent relative to the given environment action
        """
        raise NotImplementedError()

    def environment_state_to_observation(self,
                                         logger: logging.Logger,
                                         session,
                                         environment_state):
        """
        Translate the given environment state to the respective agent state, i.e. observation.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param environment_state: the state as seen by the environment
        :return: the state as seen (observed) by the agent relative to the given environment state
        """
        raise NotImplementedError()

    @property
    def observation_space_type(self) -> SpaceType:
        """
        Get the type (according to its SpaceType enum definition) of the observation space.

        :return: the SpaceType describing the observation space type
        """
        raise NotImplementedError()

    @property
    def observation_space_shape(self):
        """
        Get the shape of the observation space. This may differ from the environment state space.

        :return: the shape of the observation space in the form of a tuple
        """
        raise NotImplementedError()

    @property
    def agent_action_space_type(self) -> SpaceType:
        """
        Get the type (according to its SpaceType enum definition) of the agent action space.

        :return: the SpaceType describing the agent action space type
        """
        raise NotImplementedError()

    @property
    def agent_action_space_shape(self):
        """
        Get the shape of the agent action space. This may differ from the environment action space.

        :return: the shape of the agent action space in the form of a tuple
        """
        raise NotImplementedError()

    @property
    def environment(self):
        """
        Return the environment associated with the interface.

        :return: the environment
        """
        return self._environment



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import enum
import logging

# Define space types for observation and action spaces definition in the environment


class SpaceType(enum.Enum):
    """
    Space types for environment definition.

    If discrete, each value is one-hot encoded.
    If continuous, each value can have infinite sub values.

    Usually, the discrete space has greater size than the continuous one (it requires more variables one-hot encoded
    to properly define each state, while in continuous it is defined by all the values in the state itself).

    Except for the properties and the setup method, all the methods can be used with tensorflow by the mean of its
    session.
    """
    discrete = 1
    continuous = 2


class Environment:
    """
    Base environment abstract class.

    The environment is not created until the setup method is called.
    The environment template is inspired by the gym environment and as such should be pretty standard in a
    reinforcement learning research setting.

    To define your own environment, implement the abstract class in a specific child class.
    """

    def __init__(self,
                 name: str):
        # Define the _name attribute
        self.name: str = name

    def setup(self,
              logger: logging.Logger) -> bool:
        """
        Setup the environment.

        :param logger: the logger used to print the environment information, warnings and errors
        :return a boolean flag True if the setup is successful, false otherwise
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Initialize the environment.
        It is called right after tensorflow session generation and before the agent is initialized.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def post_initialize(self,
                        logger: logging.Logger,
                        session):
        """
        Post-initialize the environment.
        It is called after both the environment and the agent are initialized.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        """
        raise NotImplementedError()

    def close(self,
              logger: logging.Logger,
              session):
        """
        Close the environment (e.g. the window of a rendered OpenAI environment).

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def reset(self,
              logger: logging.Logger,
              session):
        """
        Reset the environment to its default state. Requires the environment to be initialized.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: the current state of the environment on its default state
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def step(self,
             logger: logging.Logger,
             action,
             session):
        """
        Execute a step on the environment with the given action. It returns back all the data describing the transition.
        Requires the environment to be initialized.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :param action: the action executed on the environment, causing the step and the state change
        :return: the next state, the reward obtained and a boolean flag indicating whether or not the episode is completed
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def render(self,
               logger: logging.Logger,
               session):
        """
        Render the environment.
        Note: to render with text, it's advised to use print statement instead of the logger.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def get_random_action(self,
                          logger: logging.Logger,
                          session):
        """
        Get a random action from the environment.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: a random action of the environment in the shape supported the environment (discrete: an int, continuous: a tuple)
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def state_space_type(self):
        """
        Get the type (according to its SpaceType enum definition) of the state space of the environment.

        :return: the SpaceType describing the state space type of the environment
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def state_space_shape(self):
        """
        Get the shape of the state space of the environment.

        :return: the shape of the state space of the environment in the form of a tuple
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def action_space_type(self):
        """
        Get the type (according to its SpaceType enum definition) of the action space of the environment.

        :return: the SpaceType describing the action space type of the environment
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def action_space_shape(self):
        """
        Get the shape of the action space of the environment.

        :return: the shape of the action space of the environment in the form of a tuple
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()





#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.po_models import ProximalPolicyOptimization


class PPOAgent(Agent):
    """
    Proximal Policy Optimization agent with Dirichlet sampling exploration policies for discrete action spaces.
    It is supplied with a PPO model. By default exploration is disabled. Change the trade-off parameters to enable it.

    The updates per training volley define how many updates should be execute by the model in each train volley. For
    example if the volley is 1000 episodes long and the number of updates per volley is 10, the agent will be updated
    every 100 episodes.

    If exploration policy is enabled, during training, agent samples from the probability distribution of its computed
    values and the dirichlet probability distribution with given alpha, according to a trade-off parameter which is
    updated at the end of each episode.
    """

    def __init__(self,
                 name: str,
                 model: ProximalPolicyOptimization,
                 updates_per_training_volley: int,
                 alpha: float = 1.0,
                 dirichlet_trade_off_min: float = 1.0, dirichlet_trade_off_max: float = 1.0,
                 dirichlet_trade_off_update: float = 0.001):
        # Define agent attributes
        self.updates_per_training_volley: int = updates_per_training_volley
        self._model: ProximalPolicyOptimization = model
        self._alpha: float = alpha
        self._dirichlet_trade_off_min: float = dirichlet_trade_off_min
        self._dirichlet_trade_off_max: float = dirichlet_trade_off_max
        self._dirichlet_trade_off_update: float = dirichlet_trade_off_update
        # Define internal agent attributes
        self._current_value_estimate = None
        self._current_log_likelihood = None
        self._current_policy_loss = None
        self._current_value_loss = None
        self._dirichlet_trade_off: float = None
        # Generate base agent
        super(PPOAgent, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape,
                  agent_action_space_type: SpaceType, agent_action_space_shape) -> bool:
        # Generate the model and check if it's successful, stop if not successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset internal agent attributes
        self._current_value_estimate = None
        self._current_log_likelihood = None
        self._current_policy_loss = None
        self._current_value_loss = None
        # Initialize the model
        self._model.initialize(logger, session)
        # Reset trade-off to its starting value (the min)
        self._dirichlet_trade_off = self._dirichlet_trade_off_min

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current):
        pass

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current):
        # Check if trade-off value is equal to 1.0:
        if self._dirichlet_trade_off >= 1.0:
            # Act according to the model best prediction (a sample from the probability distribution)
            # Note: it is still inherently exploring
            action, self._current_value_estimate, self._current_log_likelihood = self._model.sample_action(session, agent_observation_current)
        else:
            # Act according to dirichlet approach: first get probability distribution over all the actions predicted by the model
            prior_probabilities = self._model.get_action_probabilities(session, agent_observation_current)
            # Then generate a dirichlet distribution (d) with parameter alpha
            dirichlet_probabilities = numpy.random.dirichlet(self._alpha * numpy.ones(prior_probabilities.size), 1).flatten()
            # Get a random action value (random output) using x * p + (1 - x) * d as probability distribution where x is the trade-off
            output = self._dirichlet_trade_off * prior_probabilities + (1 - self._dirichlet_trade_off) * dirichlet_probabilities
            # Make sure output sums up to 1.0
            output = output / output.sum()
            action_value = numpy.random.choice(output, p=output)
            # Set the chosen action as the index of such chosen action value
            action = numpy.argmax(output == action_value)
            # Get the value and the log-likelihood of the model at the current observed state with the chosen action
            self._current_value_estimate, self._current_log_likelihood = self._model.get_value_and_log_likelihood(session, action, agent_observation_current)
        # Return the exploration action
        return action

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current):
        # Predict the action with the model by sampling from its probability distribution
        action, _, _ = self._model.sample_action(session, agent_observation_current)
        # Return the predicted action
        return action

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current,
                             agent_action,
                             reward: float,
                             agent_observation_next,
                             warmup_step_current: int,
                             warmup_episode_current: int,
                             warmup_steps_volley: int):
        pass

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current,
                            agent_action,
                            reward: float,
                            agent_observation_next,
                            train_step_current: int, train_step_absolute: int,
                            train_episode_current: int, train_episode_absolute: int,
                            train_episode_volley: int, train_episode_total: int):
        # Save the current step in the buffer together with the current value estimate
        self._model.buffer.store(agent_observation_current, agent_action, reward, self._current_value_estimate, self._current_log_likelihood)

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current,
                                agent_action,
                                reward: float,
                                agent_observation_next,
                                inference_step_current: int,
                                inference_episode_current: int,
                                inference_episode_volley: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: float,
                                episode_total_reward: float,
                                warmup_episode_current: int,
                                warmup_steps_volley: int):
        pass

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: float,
                               episode_total_reward: float,
                               train_step_absolute: int,
                               train_episode_current: int, train_episode_absolute: int,
                               train_episode_volley: int, train_episode_total: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_path(last_step_reward)
        # Increase trade-off by its update value
        self._dirichlet_trade_off = min(self._dirichlet_trade_off_max, self._dirichlet_trade_off + self._dirichlet_trade_off_update)
        # Execute update only after a certain number of trajectories each time (need to offset the episode by 1 to prevent not updating last episode)
        if (train_episode_current + 1) % (train_episode_volley / self.updates_per_training_volley) == 0 and train_episode_current > 0:
            # Execute the update and store policy and value loss
            buffer: [] = self._model.buffer.get()
            logger.info("Updating the model with a buffer of " + str(len(buffer[0])) + " samples")
            summary, self._current_policy_loss, self._current_value_loss = self._model.update(session, buffer)
            # Update the summary at the absolute current step
            self._summary_writer.add_summary(summary, train_step_absolute)

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: float,
                                   episode_total_reward: float,
                                   inference_episode_current: int,
                                   inference_episode_volley: int):
        pass

    @property
    def trainable_variables(self):
        # Return the trainable variables of the agent model in experiment/agent _scope
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        # Return the amount of warmup episodes required by the model
        return self._model.warmup_steps






# Import packages

import tensorflow
import numpy
import scipy.signal
import random

# Import required src

from usienarl import SpaceType, Config, Model
from usienarl.utils.common import softmax


class Buffer:
    """
    A buffer for storing trajectories experienced by a PPO agent interacting with the environment,
    and using Generalized Advantage Estimation (GAE-Lambda) for calculating the advantages of state-action pairs.

    The buffer is dynamically resizable.

    The buffer contains list of states (or observations), actions (used as targets), values (computed by the value stream
    of the _model itself during prediction), advantages (computed by the buffer using GAE when a trajectory finishes and
    fed back up in the policy stream to drive the loss), rewards (used to compute rewards-to-go), rewards-to-go
    (computed inside the buffer itself and used as weight for the targets action when training the value stream) and
    log-likelihoods (used to compute policy ratio).
    """

    def __init__(self,
                 discount_factor: float, lambda_parameter: float):
        # Define buffer components
        self._observations: [] = []
        self._actions: [] = []
        self._advantages: [] = []
        self._rewards: [] = []
        self._rewards_to_go: [] = []
        self._values: [] = []
        self._log_likelihoods: [] = []
        # Define parameters
        self._discount_factor: float = discount_factor
        self._lambda_parameter: float = lambda_parameter
        # Define buffer pointer
        self._pointer: int = 0
        self._path_start_index: int = 0

    def store(self,
              observation, action, reward: float, value: float, log_likelihood):
        """
        Store the time-step in the buffer.

        :param observation: the current observation to store in the buffer
        :param action: the last action to store in the buffer
        :param reward: the reward obtained from the action at the current state to store in the buffer
        :param value: the value of the state as estimated by the value stream of the model to store in the buffer
        :param log_likelihood: the log likelihood of the action on the state as estimated by the model
        """
        # Append all data and increase the pointer
        self._observations.append(observation)
        self._actions.append(action)
        self._rewards.append(reward)
        self._values.append(value)
        self._log_likelihoods.append(log_likelihood)
        self._pointer += 1

    def get(self) -> []:
        """
        Get all of the data from the buffer, with advantages appropriately normalized (shifted to have mean zero and
        standard deviation equals to one). Also reset pointers in the buffer and the lists composing the buffer.

        :return a list containing the ndarrays of: states, actions, advantages, rewards-to-go, log-likelihoods
        """
        # Get a numpy array on the advantage list
        advantages_array: numpy.ndarray = numpy.array(self._advantages)
        # Execute the advantage normalization trick
        # Note: make sure mean and std are not zero!
        global_sum: float = numpy.sum(advantages_array)
        advantage_mean: float = global_sum / advantages_array.size + 1e-8
        global_sum_squared: float = numpy.sum((advantages_array - advantage_mean) ** 2) + 1e-8
        advantage_std: float = numpy.sqrt(global_sum_squared / advantages_array.size) + 1e-8
        # Adjust advantages according to the trick
        advantages_array = ((advantages_array - advantage_mean) / advantage_std)
        # Save the necessary values as numpy arrays before reset
        states_array: numpy.ndarray = numpy.array(self._observations)
        actions_array: numpy.ndarray = numpy.array(self._actions)
        rewards_to_go_array: numpy.ndarray = numpy.array(self._rewards_to_go)
        log_likelihoods_array: numpy.ndarray = numpy.array(self._log_likelihoods)
        # Reset the buffer and related pointers
        self._pointer = 0
        self._path_start_index = 0
        self._observations = []
        self._actions = []
        self._advantages = []
        self._rewards = []
        self._rewards_to_go = []
        self._values = []
        self._log_likelihoods = []
        # Return all the buffer components
        return [states_array, actions_array, advantages_array, rewards_to_go_array, log_likelihoods_array]

    def finish_path(self,
                    value: float = 0):
        """
        Finish the path at the end of a trajectory. This looks back in the buffer to where the trajectory started,
        and uses rewards and value estimates from the whole trajectory to compute advantage estimates with GAE-Lambda,
        as well as compute the rewards-to-go for each state, to use as the targets for the value stream optimization.

        :param value: the last reward given by the environment or the last predicted value if last state is not terminal
        """
        path_slice = slice(self._path_start_index, self._pointer)
        rewards: numpy.ndarray = numpy.array(self._rewards[path_slice] + [value])
        values: numpy.ndarray = numpy.array(self._values[path_slice] + [value])
        # Compute GAE-Lambda advantage estimation (compute advantages using the value in the buffer taken from the model)
        deltas: numpy.ndarray = rewards[:-1] + self._discount_factor * values[1:] - values[:-1]
        self._advantages[path_slice] = self._discount_cumulative_sum(deltas, self._discount_factor * self._lambda_parameter).tolist()
        # Compute rewards-to-go
        self._rewards_to_go[path_slice] = (self._discount_cumulative_sum(rewards, self._discount_factor)[:-1]).tolist()
        self._path_start_index = self._pointer

    @property
    def size(self) -> int:
        """
        The size of the buffer at the current time (it is dynamic).

        :return: the integer size of the buffer.
        """
        return self._pointer

    @staticmethod
    def _discount_cumulative_sum(vector: numpy.ndarray, discount: float) -> numpy.ndarray:
        """
        Compute discounted cumulative sums of vectors.
        Credits to rllab.

        :param vector: the vector on which to compute cumulative discounted sum (e.g. [x0, x1, x2])
        :return the discounted cumulative sum (e.g. [x0 + discount * x1 + discount^2 * x2, x1 + discount * x2, x3])
        """
        return scipy.signal.lfilter([1], [1, float(-discount)], vector[::-1], axis=0)[::-1]


class ProximalPolicyOptimization(Model):
    """
    Proximal Policy Optimization with GAE (Generalized Advantage Estimation).
    The algorithm is on-policy and executes updates every a certain number of episodes.
    The model is constituted by two sub-models, or streams. The first stream computes and optimizes the policy loss,
    and to drive the loss the advantages for each state in the batch is required to be estimated. The loss used an
    approximation of the KL divergence. This stream is called policy stream.
    The second stream computes the value on the current states and optimizes such estimation. To drive the loss of such
    sub-model, the value estimated by the stream itself for each state in the batch is required to be estimated. This
    stream is called value stream.
    The advantage used to drive the policy stream loss is computed using GAE on the value estimated by the value stream
    and such computation is carried on in the buffer.

    The buffer stores all the trajectories up to the update point. Since each episode can contains different
    numbers of steps, the buffer is dynamically resizable.
    The algorithm is very likely to converge to local minima but guarantees to not decrease its policy quality according
    to a minimum KL distance allowed between old and new policy.

    Supported observation spaces:
        - discrete
        - continuous

    Supported action spaces:
        - discrete
        - continuous
    """

    def __init__(self,
                 name: str,
                 discount_factor: float,
                 learning_rate_policy: float, learning_rate_value: float,
                 value_steps_for_update: int, policy_steps_for_update: int,
                 minibatch_size: int,
                 hidden_layers_config: Config,
                 lambda_parameter: float,
                 clip_ratio: float,
                 target_kl_divergence: float):
        # Define model attributes
        self.learning_rate_policy: float = learning_rate_policy
        self.learning_rate_value: float = learning_rate_value
        self.discount_factor: float = discount_factor
        # Define internal model attributes
        self._hidden_layers_config: Config = hidden_layers_config
        self._value_steps_for_update: int = value_steps_for_update
        self._policy_steps_for_update: int = policy_steps_for_update
        self._minibatch_size: int = minibatch_size
        self._lambda_parameter: float = lambda_parameter
        self._clip_ratio: float = clip_ratio
        self._target_kl_divergence: float = target_kl_divergence
        # Define proximal policy optimization empty attributes
        self.buffer: Buffer = None
        # Define internal proximal policy optimization empty attributes
        self._inputs = None
        self._actions = None
        self._targets = None
        self._advantages = None
        self._rewards = None
        self._mask = None
        self._logits = None
        self._masked_logits = None
        self._expected_value = None
        self._std = None
        self._log_std = None
        self._log_likelihood_targets = None
        self._log_likelihood_actions = None
        self._previous_log_likelihoods = None
        self._value = None
        self._ratio = None
        self._min_advantage = None
        self._value_stream_loss = None
        self._policy_stream_loss = None
        self._value_stream_optimizer = None
        self._policy_stream_optimizer = None
        self._approximated_kl_divergence = None
        self._approximated_entropy = None
        self._clip_fraction = None
        # Generate the base model
        super(ProximalPolicyOptimization, self).__init__(name)
        # Define the types of allowed observation and action spaces
        self._supported_observation_space_types.append(SpaceType.discrete)
        self._supported_observation_space_types.append(SpaceType.continuous)
        self._supported_action_space_types.append(SpaceType.discrete)
        self._supported_action_space_types.append(SpaceType.continuous)

    def _define_graph(self):
        # Set the GAE buffer for the vanilla policy gradient algorithm
        self.buffer: Buffer = Buffer(self.discount_factor, self._lambda_parameter)
        # Define the tensorflow model
        with tensorflow.variable_scope(self._scope + "/" + self._name):
            # Define inputs of the estimator as a float adaptable array with shape Nx(S) where N is the number of examples and (S) the shape of the state
            self._inputs = tensorflow.placeholder(shape=[None, *self._observation_space_shape], dtype=tensorflow.float32, name="inputs")
            # Define the estimator network hidden layers from the config
            hidden_layers_output = self._hidden_layers_config.apply_hidden_layers(self._inputs)
            # Define the targets for learning with the same NxA adaptable size
            self._targets = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="targets")
            # Change the _model definition according to its action space type
            if self._agent_action_space_type == SpaceType.discrete:
                # Define the mask placeholder
                self._mask = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="mask")
                # Define the logits as outputs of the deep neural network with shape NxA where N is the number of inputs, A is the action size when its type is discrete
                self._logits = tensorflow.layers.dense(hidden_layers_output, *self._agent_action_space_shape, activation=None, kernel_initializer=tensorflow.contrib.layers.xavier_initializer(), name="logits")
                # Compute the masked logits using the given mask
                self._masked_logits = tensorflow.add(self._logits, self._mask)
                # Define the actions on the first shape dimension as a squeeze on the samples drawn from a categorical distribution on the logits
                self._actions = tensorflow.squeeze(tensorflow.multinomial(logits=self._masked_logits, num_samples=1), axis=1)
                # Define the log likelihood according to the categorical distribution on targets and actions
                self._log_likelihood_targets, _ = self.get_categorical_log_likelihood(self._targets, self._logits)
                self._log_likelihood_actions, _ = self.get_categorical_log_likelihood(tensorflow.one_hot(self._actions, depth=self._agent_action_space_shape[0]), self._logits)
            else:
                # Define the expected value as the output of the deep neural network with shape Nx(A) where N is the number of inputs, (A) is the action shape
                self._expected_value = tensorflow.layers.dense(hidden_layers_output, *self._agent_action_space_shape, activation=None, kernel_initializer=tensorflow.contrib.layers.xavier_initializer(), name="expected_value")
                # Define the log standard deviation
                self._log_std = tensorflow.get_variable(name="log_std", initializer=-0.5*numpy.ones(*self._agent_action_space_shape, dtype=numpy.float32))
                # Define the standard deviation
                self._std = tensorflow.exp(self._log_std, name="std")
                # Define actions as the expected value summed up with a noise vector multiplied by the standard deviation
                self._actions = self._expected_value + tensorflow.random_normal(tensorflow.shape(self._expected_value)) * self._std
                # Define the log likelihood according to the gaussian distribution on targets and actions
                self._log_likelihood_targets = self.get_gaussian_log_likelihood(self._targets, self._expected_value, self._log_std)
                self._log_likelihood_actions = self.get_gaussian_log_likelihood(self._actions, self._expected_value, self._log_std)
            # Define the value estimator (a deep MLP)
            value_stream_hidden_layers_output = self._hidden_layers_config.apply_hidden_layers(self._inputs)
            value_stream_output = tensorflow.layers.dense(value_stream_hidden_layers_output, 1, activation=None, kernel_initializer=tensorflow.contrib.layers.xavier_initializer(), name="value")
            # Define value by squeezing the output of the advantage stream MLP
            self._value = tensorflow.squeeze(value_stream_output, axis=1, name="value")
            # Define the rewards as an adaptable vector of floats (they are actually rewards-to-go computed with GAE)
            self._rewards = tensorflow.placeholder(shape=(None, ), dtype=tensorflow.float32, name="rewards")
            # Define advantage loss as the mean squared error of the difference between computed rewards to go and the advantage
            self._value_stream_loss = tensorflow.reduce_mean((self._rewards - self._value) ** 2, name="value_loss")
            # Define the optimizer for the value stream (actually the MLP optimizer)
            self._value_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_value).minimize(self._value_stream_loss)
            # Define the advantages placeholder as an adaptable vector of floats (they are stored in the buffer)
            # Note: the model get the advantages from the buffer once computed using GAE on the values
            self._advantages = tensorflow.placeholder(shape=(None,), dtype=tensorflow.float32, name="advantages")
            # Define the previous log-likelihood placeholder as an adaptable vector or float (they are stored in the buffer)
            # Note: the model get the previous log-likelihoods from the buffer
            self._previous_log_likelihoods = tensorflow.placeholder(shape=(None,), dtype=tensorflow.float32, name="previous_log_likelihood")
            # Define the ratio between the current log-likelihood and the previous one (when using exponential, minus is a division)
            self._ratio = tensorflow.exp(self._log_likelihood_targets - self._previous_log_likelihoods)
            # Define the minimum advantages with respect to clip ratio
            self._min_advantage = tensorflow.where(self._advantages > 0, (1 + self._clip_ratio) * self._advantages, (1 - self._clip_ratio) * self._advantages)
            # Define the policy stream loss as the mean of minimum between the advantages multiplied the ratio and the minimum advantage
            self._policy_stream_loss = -tensorflow.reduce_mean(tensorflow.minimum(self._ratio * self._advantages, self._min_advantage), name="policy_loss")
            # Define the optimizer for the policy stream
            self._policy_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_policy).minimize(self._policy_stream_loss)
            # Define approximated KL divergence (also used to early stop), approximated entropy and clip fraction for the logger
            self._approximated_kl_divergence = tensorflow.reduce_mean(self._previous_log_likelihoods - self._log_likelihood_targets, name="approximated_kl_divergence")
            self._approximated_entropy = tensorflow.reduce_mean(-self._log_likelihood_targets, name="approximated_entropy")
            self._clip_fraction = tensorflow.reduce_mean(tensorflow.cast(tensorflow.logical_or(self._ratio > (1 + self._clip_ratio), self._ratio < (1 - self._clip_ratio)), tensorflow.float32), name="clip_fraction")
            # Define the initializer
            self._initializer = tensorflow.global_variables_initializer()

    def sample_action(self,
                      session,
                      observation_current,
                      mask: numpy.ndarray = None):
        """
        Get the action sampled from the probability distribution of the model given the current observation and an optional mask.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon
        :param mask: the optional mask used (only if the action space type is discrete) to remove certain actions from the prediction (-infinity to remove, 0.0 to pass-through)
        :return: the action predicted by the model, with the value estimated at the current state and the relative log-likelihood of the sampled action
        """
        # If there is no mask and the action space type is discrete generate a full pass-through mask
        if mask is None and self._agent_action_space_type == SpaceType.discrete:
            mask = numpy.zeros(self._agent_action_space_shape, dtype=float)
        # Return a random action sample given the current state and depending on the observation space type and also compute value estimate and log-likelihood
        if self._observation_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the observation if observation space type is discrete
            observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
            if self._agent_action_space_type == SpaceType.discrete:
                action, value, log_likelihood = session.run([self._actions, self._value, self._log_likelihood_actions],
                                                            feed_dict={self._inputs: [observation_current_one_hot], self._mask: [mask]})
            else:
                action, value, log_likelihood = session.run([self._actions, self._value, self._log_likelihood_actions],
                                                            feed_dict={self._inputs: [observation_current_one_hot]})
        else:
            if self._agent_action_space_type == SpaceType.discrete:
                action, value, log_likelihood = session.run([self._actions, self._value, self._log_likelihood_actions],
                                                            feed_dict={self._inputs: [observation_current], self._mask: [mask]})
            else:
                action, value, log_likelihood = session.run([self._actions, self._value, self._log_likelihood_actions],
                                                            feed_dict={self._inputs: [observation_current]})
        # Return the predicted action, the estimated value and the log-likelihood
        return action[0], value[0], log_likelihood[0]

    def get_value_and_log_likelihood(self,
                                     session,
                                     action,
                                     observation_current,
                                     mask: numpy.ndarray = None):
        """
        Get the estimated value of the given current observation and the log-likelihood of the given action.

        :param session: the session of tensorflow currently running
        :param action: the action of which to compute the log-likelihood
        :param observation_current: the current observation of the agent in the environment to estimate the value
        :param mask: the optional mask used (only if the action space type is discrete) to remove certain actions from the prediction (-infinity to remove, 0.0 to pass-through)
        :return: the value estimated at the current state and the log-likelihood of the given action
        """
        # If there is no mask and the action space type is discrete generate a full pass-through mask
        if mask is None and self._agent_action_space_type == SpaceType.discrete:
            mask = numpy.zeros(self._agent_action_space_shape, dtype=float)
        # Generate a one-hot encoded version of the action if action space type is discrete
        if self._agent_action_space_type == SpaceType.discrete:
            action: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[numpy.array(action).reshape(-1)]
        # Return the estimated value of the given current state and the log-likelihood of the given action
        if self._observation_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the observation if observation space type is discrete
            observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
            if self._agent_action_space_type == SpaceType.discrete:
                value, log_likelihood = session.run([self._value, self._log_likelihood_targets],
                                                    feed_dict={self._inputs: [observation_current_one_hot],
                                                               self._mask: [mask],
                                                               self._targets: action})
            else:
                value, log_likelihood = session.run([self._value, self._log_likelihood_targets],
                                                    feed_dict={self._inputs: [observation_current_one_hot],
                                                               self._targets: action})
        else:
            if self._agent_action_space_type == SpaceType.discrete:
                value, log_likelihood = session.run([self._value, self._log_likelihood_targets],
                                                    feed_dict={self._inputs: [observation_current],
                                                               self._mask: [mask],
                                                               self._targets: action})
            else:
                value, log_likelihood = session.run([self._value, self._log_likelihood_targets],
                                                    feed_dict={self._inputs: [observation_current],
                                                              self._targets: action})
        # Return the estimated value and the log-likelihood
        return value[0], log_likelihood[0]

    def get_value(self,
                  session,
                  observation_current):
        """
        Get the estimated value of the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to estimate the value
        :return: the value estimated at the current state
        """
        # Return the value predicted by the network at the current state
        if self._observation_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the observation if observation space type is discrete
            observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
            value = session.run(self._value,
                                feed_dict={self._inputs: [observation_current_one_hot]})
        else:
            value = session.run(self._value,
                                feed_dict={self._inputs: [observation_current]})
        # Return the estimated value
        return value[0]

    def get_log_likelihood(self,
                           session,
                           action,
                           observation_current,
                           mask: numpy.ndarray = None):
        """
        Get the the log-likelihood of the given action.

        :param session: the session of tensorflow currently running
        :param action: the action of which to compute the log-likelihood
        :param observation_current: the current observation of the agent in the environment
        :param mask: the optional mask used (only if the action space type is discrete) to remove certain actions from the prediction (-infinity to remove, 0.0 to pass-through)
        :return: the log-likelihood of the given action
        """
        # If there is no mask and the action space type is discrete generate a full pass-through mask
        if mask is None and self._agent_action_space_type == SpaceType.discrete:
            mask = numpy.zeros(self._agent_action_space_shape, dtype=float)
        # Generate a one-hot encoded version of the action if action space type is discrete
        if self._agent_action_space_type == SpaceType.discrete:
            action: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[numpy.array(action).reshape(-1)]
        # Return log likelihood over the given action with the given current observation and the given mask
        if self._observation_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the observation if observation space type is discrete
            observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
            if self._agent_action_space_type == SpaceType.discrete:
                log_likelihood = session.run(self._log_likelihood_targets,
                                             feed_dict={self._inputs: [observation_current_one_hot],
                                                        self._mask: [mask],
                                                        self._targets: action})
            else:
                log_likelihood = session.run(self._log_likelihood_targets,
                                             feed_dict={self._inputs: [observation_current_one_hot],
                                                        self._targets: action})
        else:
            if self._agent_action_space_type == SpaceType.discrete:
                log_likelihood = session.run(self._log_likelihood_targets,
                                             feed_dict={self._inputs: [observation_current],
                                                        self._mask: [mask],
                                                        self._targets: action})
            else:
                log_likelihood = session.run(self._log_likelihood_targets,
                                             feed_dict={self._inputs: [observation_current],
                                                        self._targets: action})
        # Return the log-likelihood
        return log_likelihood[0]

    def get_action_probabilities(self,
                                 session,
                                 observation_current,
                                 mask: numpy.ndarray = None) -> []:
        """
        Get all the action probabilities (softmax over masked logits if discrete, expected value if continuous) for the
        given current observation and an optional mask.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon
        :param mask: the optional mask used (only if the action space type is discrete) to remove certain actions from the prediction (-infinity to remove, 0.0 to pass-through)
        :return: the list of action probabilities (softmax over masked logits or expected values depending on the agent action space type)
        """
        if mask is None and self._agent_action_space_type == SpaceType.discrete:
            mask = numpy.zeros(self._agent_action_space_shape, dtype=float)
        # Get the logits or the expected value as the distribution of the action probabilities depending on the action space shape
        if self._agent_action_space_type == SpaceType.discrete:
            if self._observation_space_type == SpaceType.discrete:
                observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
                if self._agent_action_space_type == SpaceType.discrete:
                    logits = session.run([self._masked_logits],
                                         feed_dict={self._inputs: [observation_current_one_hot], self._mask: [mask]})
                else:
                    logits = session.run([self._masked_logits],
                                         feed_dict={self._inputs: [observation_current_one_hot]})
            else:
                if self._agent_action_space_type == SpaceType.discrete:
                    logits = session.run([self._masked_logits],
                                         feed_dict={self._inputs: [observation_current], self._mask: [mask]})
                else:
                    logits = session.run([self._masked_logits],
                                         feed_dict={self._inputs: [observation_current]})
            # Return the softmax over the logits (probabilities of all actions)
            return softmax(logits[0]).flatten()
        else:
            expected_value, std = session.run([self._expected_value, self._std],
                                              feed_dict={self._inputs: [observation_current]})
            # Return the expected value
            return expected_value

    def update(self,
               session,
               batch: []):
        # Unpack the batch to feed the minibatches
        observations, actions, advantages, rewards, previous_log_likelihoods = batch[0], batch[1], batch[2], batch[3], batch[4]
        # Generate a one-hot encoded version of the observations if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observations: numpy.ndarray = numpy.eye(*self._observation_space_shape)[numpy.array(observations).reshape(-1)]
        # Generate a one-hot encoded version of the actions if action space type is discrete
        if self._agent_action_space_type == SpaceType.discrete:
            actions: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[numpy.array(actions).reshape(-1)]
        # Run the policy optimizer of the model in training mode for the required amount of steps, also compute policy stream loss, approximated kl divergence, approximated entropy and clip fraction
        policy_stream_loss_average: float = 0.0
        approximated_kl_divergence_average: float = 0.0
        approximated_entropy_average: float = 0.0
        clip_fraction_average: float = 0.0
        for _ in range(self._policy_steps_for_update):
            policy_update_minibatch_iterations: int = 0
            policy_stream_loss_total: float = 0.0
            approximated_kl_divergence_total: float = 0.0
            approximated_entropy_total: float = 0.0
            clip_fraction_total: float = 0.0
            for minibatch in self._get_minibatch(observations, actions, advantages, rewards, previous_log_likelihoods, self._minibatch_size):
                # Unpack the minibatch
                minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards, minibatch_previous_log_likelihoods = minibatch
                # Update the policy
                _, policy_stream_loss, approximated_kl_divergence, approximated_entropy, clip_fraction = session.run([self._policy_stream_optimizer, self._policy_stream_loss, self._approximated_kl_divergence, self._approximated_entropy, self._clip_fraction],
                                                                                                                     feed_dict={
                                                                                                                         self._inputs: minibatch_observations,
                                                                                                                         self._targets: minibatch_actions,
                                                                                                                         self._advantages: minibatch_advantages,
                                                                                                                         self._previous_log_likelihoods: minibatch_previous_log_likelihoods
                                                                                                                     })
                policy_update_minibatch_iterations += 1
                policy_stream_loss_total += policy_stream_loss
                approximated_kl_divergence_total += approximated_kl_divergence
                approximated_entropy_total += approximated_entropy
                clip_fraction_total += clip_fraction
            # The average is only really saved on the last value update to know it at the end of all the update steps
            policy_stream_loss_average = policy_stream_loss_total / policy_update_minibatch_iterations
            approximated_kl_divergence_average = approximated_kl_divergence_total / policy_update_minibatch_iterations
            approximated_entropy_average = approximated_entropy_total / policy_update_minibatch_iterations
            clip_fraction_average = clip_fraction_total / policy_update_minibatch_iterations
            # If average approximated KL divergence over last policy update step (all minibatches) is above a certain threshold stop updating the policy for now (early stop)
            if approximated_kl_divergence_average > 1.5 * self._target_kl_divergence:
                break
        # Run the value optimizer of the model in training mode for the required amount of steps, also compute average value stream loss
        value_stream_loss_average: float = 0.0
        for _ in range(self._value_steps_for_update):
            value_update_minibatch_iterations: int = 0
            value_stream_loss_total: float = 0.0
            for minibatch in self._get_minibatch(observations, actions, advantages, rewards, previous_log_likelihoods, self._minibatch_size):
                # Unpack the minibatch
                minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards, minibatch_previous_log_likelihoods = minibatch
                # Update the value
                _, value_stream_loss = session.run([self._value_stream_optimizer, self._value_stream_loss],
                                                   feed_dict={
                                                        self._inputs: minibatch_observations,
                                                        self._advantages: minibatch_advantages,
                                                        self._rewards: minibatch_rewards
                                                    })
                value_update_minibatch_iterations += 1
                value_stream_loss_total += value_stream_loss
            # The average is only really saved on the last value update to know it at the end of all the update steps
            value_stream_loss_average = value_stream_loss_total / value_update_minibatch_iterations
        # Generate the tensorflow summary
        summary = tensorflow.Summary()
        summary.value.add(tag="policy_stream_loss", simple_value=policy_stream_loss_average)
        summary.value.add(tag="value_stream_loss", simple_value=value_stream_loss_average)
        summary.value.add(tag="approximated_kl_divergence", simple_value=approximated_kl_divergence_average)
        summary.value.add(tag="approximated_entropy", simple_value=approximated_entropy_average)
        summary.value.add(tag="clip_fraction", simple_value=clip_fraction_average)
        # Return both losses and summary for the update sequence
        return summary, policy_stream_loss_average, value_stream_loss_average

    @property
    def warmup_steps(self) -> int:
        return 0

    @staticmethod
    def _get_minibatch(observations: [], actions: [], advantages: [], rewards: [], previous_log_likelihoods: [],
                       minibatch_size: int) -> ():
        """
        Get a minibatch of the given minibatch size from the given batch (already unpacked).

        :param observations: the observations buffer in the batch
        :param actions: the actions buffer in the batch
        :param advantages: the advantages buffer in the batch
        :param rewards: the rewards buffer in the batch
        :param previous_log_likelihoods: the previous log-likelihoods buffer in the batch
        :param minibatch_size: the size of the minibatch
        :return: a tuple minibatch of shuffled samples of the given size
        """
        # Get a list of random ids of the batch
        batch_random_ids = random.sample(range(len(observations)), len(observations))
        # Generate the minibatches by shuffling the batch
        minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards, minibatch_previous_log_likelihoods = [], [], [], [], []
        for random_id in batch_random_ids:
            minibatch_observations.append(observations[random_id])
            minibatch_actions.append(actions[random_id])
            minibatch_advantages.append(advantages[random_id])
            minibatch_rewards.append(rewards[random_id])
            minibatch_previous_log_likelihoods.append(previous_log_likelihoods[random_id])
            # Return the minibatch
            if len(minibatch_observations) % minibatch_size == 0:
                yield minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards,  minibatch_previous_log_likelihoods
                # Clear the minibatch
                minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards, minibatch_previous_log_likelihoods = [], [], [], [], []

    @staticmethod
    def get_categorical_log_likelihood(actions_mask, logits):
        """
        Get log-likelihood for discrete action spaces (using a categorical distribution) on the given logits and with
        the action mask.
        It uses tensorflow and as such should only be called in the define method.

        :param actions_mask: the actions used to mask the log-likelihood on the logits
        :param logits: the logits of the neural network
        :return: the log-likelihood according to categorical distribution
        """
        # Define the unmasked likelihood as the log-softmax of the logits
        log_likelihood_unmasked = tensorflow.nn.log_softmax(logits)
        # Return the categorical log-likelihood by summing over the first axis of the actions mask multiplied
        # by the log-likelihood on the logits (unmasked, this is the masking operation) and the unmasked likelihood
        return tensorflow.reduce_sum(actions_mask * log_likelihood_unmasked, axis=1, name="log_likelihood"), log_likelihood_unmasked

    @staticmethod
    def get_gaussian_log_likelihood(actions, expected_value, log_std):
        """
        Get log-likelihood for continuous action spaces (using a gaussian distribution) on the given expected value and
        log-std and with the actions.
        It uses tensorflow and as such should only be called in the define method.

        :param actions: the actions used to compute the log-likelihood tensor
        :param expected_value: the expected value of the gaussian distribution
        :param log_std: the log-std of the gaussian distribution
        :return: the log-likelihood according to gaussian distribution
        """
        # Define the log-likelihood tensor for the gaussian distribution on the given actions
        log_likelihood_tensor = -0.5 * (((actions - expected_value) / (tensorflow.exp(log_std) + 1e-8)) ** 2 + 2 * log_std + numpy.log(2 * numpy.pi))
        # Return the gaussian log-likelihood by summing over all the elements in the log-likelihood tensor defined above
        return tensorflow.reduce_sum(log_likelihood_tensor, axis=1, name="log_likelihood")



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import tensorflow
import enum


# Define layer types for tensorflow neural networks definition

class LayerType(enum.Enum):
    """
    Layer types for layer definition.

    Each layer type follows the naming scheme of the relative tensorflow layer types.
    Choosing a layer type actually will make the _model use the relative tensorflow layer.
    """
    average_pooling_1D = 1
    average_pooling_2D = 2
    average_pooling_3D = 3
    batch_normalization = 4
    convolution_1D = 5
    convolution_2D = 6
    convolution_2D_transpose = 7
    convolution_3D = 8
    convolution_3D_transpose = 9
    dense = 10
    dropout = 11
    flatten = 12
    max_pooling_1D = 13
    max_pooling_2D = 14
    max_pooling_3D = 15
    separable_convolution_1D = 16
    separable_convolution_2D = 17


class Config:
    """
    Wrapper class used to define neural networks.

    Parts (like hidden layers) of a tensorflow neural network model can be defined by it.
    """

    def __init__(self):
        self._hidden_layers_config: [] = []

    def add_hidden_layer(self,
                         layer_type, layer_parameters: []):
        """
        Add an hidden layer to the configuration using a tensorflow layer type as defined in the LayerType enum or as
        a custom defined string and a list of ordered layer parameters according to its specification.

        :param layer_type: the LayerType enum of the layer (following tensorflow names) or a custom string
        :param layer_parameters: the ordered list of parameters of the layer
        """
        # Add the layer type to this hidden layer definition list
        hidden_layer_config: [] = [layer_type]
        # Add each parameter to this hidden layer definition list
        for layer_parameter in layer_parameters:
            hidden_layer_config.append(layer_parameter)
        # Add this hidden layer definition list to the hidden layers definition list
        self._hidden_layers_config.append(hidden_layer_config)

    def apply_hidden_layers(self,
                            input_layer):
        """
        Apply the hidden layer to the network on the given input layer.

        :param input_layer: the input layer of the network
        :return: the output of the last hidden layer, usually connected to the neural network outputs.
        """
        # Initialize a list of hidden layers (to compute input/outputs)
        hidden_layers = []
        # Define all layers according to defined hidden layers type and parameters
        for index, hidden_layer_config in enumerate(self._hidden_layers_config):
            # Use layer type and parameters to define each layer
            layer_type = hidden_layer_config[0]
            layer_parameters: [] = hidden_layer_config[1:]
            # Define the input for the current layer
            hidden_layer_input = input_layer
            if isinstance(layer_type, str):
                # Custom hidden layer of name layer type
                hidden_layer = self._define_custom_hidden_layers(layer_type, hidden_layer_input, layer_parameters)
            else:
                if index > 0:
                    hidden_layer_input = hidden_layers[index - 1]
                # Define the current layer depending on the specified type and parameters
                if layer_type == LayerType.average_pooling_1D:
                    hidden_layer = tensorflow.layers.average_pooling1d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.average_pooling_2D:
                    hidden_layer = tensorflow.layers.average_pooling2d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.average_pooling_3D:
                    hidden_layer = tensorflow.layers.average_pooling3d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.batch_normalization:
                    hidden_layer = tensorflow.layers.batch_normalization(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.convolution_1D:
                    hidden_layer = tensorflow.layers.conv1d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.convolution_2D:
                    hidden_layer = tensorflow.layers.conv2d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.convolution_2D_transpose:
                    hidden_layer = tensorflow.layers.conv2d_transpose(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.convolution_3D:
                    hidden_layer = tensorflow.layers.conv3d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.convolution_3D_transpose:
                    hidden_layer = tensorflow.layers.conv3d_transpose(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.separable_convolution_1D:
                    hidden_layer = tensorflow.layers.separable_conv1d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.separable_convolution_2D:
                    hidden_layer = tensorflow.layers.separable_conv2d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.dense:
                    hidden_layer = tensorflow.layers.dense(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.dropout:
                    hidden_layer = tensorflow.layers.dropout(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.flatten:
                    hidden_layer = tensorflow.layers.flatten(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.max_pooling_1D:
                    hidden_layer = tensorflow.layers.max_pooling1d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.max_pooling_2D:
                    hidden_layer = tensorflow.layers.max_pooling2d(hidden_layer_input, *layer_parameters)
                elif layer_type == LayerType.max_pooling_3D:
                    hidden_layer = tensorflow.layers.max_pooling3d(hidden_layer_input, *layer_parameters)
                # An error occurred, break
                else:
                    return
            # Add the last defined layer in the layers list
            hidden_layers.append(hidden_layer)
        # Return the last layer (to compute the output)
        return hidden_layers[-1]

    def _define_custom_hidden_layers(self, layer_name: str, layer_input, layer_parameters: []):
        """
        Define custom hidden layers indexed by the given layer name string. These layers take as input the given
        layer input, usually a tensor, and as parameters the given ordered list of parameters.

        :param layer_name: the string name of the layers
        :param layer_input: the input of the hidden layer (usually when applied the output of the layer before)
        :param layer_parameters: the ordered list of parameters of the layer (the same of add_custom_hidden_layer)
        :return: the output of the custom hidden layer (usually when applied what is fed to the following layer)
        """
        raise NotImplementedError()




#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy
import random

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.td_models import TabularSARSA


class TabularSARSAAgentEpsilonGreedy(Agent):
    """
    Tabular SARSA agent with epsilon greedy exploration policy.
    It is supplied with a Tabular SARSA model and an exploration policy.

    The batch size define how many steps from the prioritized experience replay built-in buffer should be fed into
    the model when updating. The agent uses a 1-step temporal difference algorithm, i.e. it is updated at every step.

    During training, agent will chose a random action with probability epsilon. Such probability is called exploration
    rate and decays at every completed episode of its decay value, while remaining bounded by its defined max and min
    values.
    """

    def __init__(self,
                 name: str,
                 model: TabularSARSA,
                 batch_size: int = 1,
                 exploration_rate_max: float = 1.0, exploration_rate_min: float = 0.001,
                 exploration_rate_decay: float = 0.001):
        # Define agent attributes
        self._model: TabularSARSA = model
        self._exploration_rate_max: float = exploration_rate_max
        self._exploration_rate_min: float = exploration_rate_min
        self._exploration_rate_decay: float = exploration_rate_decay
        # Define internal agent attributes
        self._batch_size: int = batch_size
        self._current_absolute_errors = None
        self._current_loss = None
        self._next_warmup_action = None
        self._next_train_action = None
        self._exploration_rate: float = None
        # Generate base agent
        super(TabularSARSAAgentEpsilonGreedy, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape,
                  agent_action_space_type: SpaceType, agent_action_space_shape) -> bool:
        # Generate the model and return a flag stating if generation was successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset internal agent attributes
        self._current_absolute_errors = None
        self._current_loss = None
        self._next_warmup_action = None
        self._next_train_action = None
        # Initialize the model
        self._model.initialize(logger, session)
        # Reset exploration rate to its starting value (the max)
        self._exploration_rate = self._exploration_rate_max

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current):
        # Act randomly or with the already chosen random action, if any
        action = self._next_warmup_action
        if self._next_warmup_action is None:
            action = interface.get_random_agent_action(logger, session)
        # Return the random action
        return action

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current):
        # Get the best action predicted by the model or use the already predicted action, if any
        action = self._next_train_action
        if self._next_train_action is None:
            # Choose an action according to the epsilon greedy approach: best action predicted by the model or random action
            if self._exploration_rate > 0 and random.uniform(0, 1.0) < self._exploration_rate:
                action = interface.get_random_agent_action(logger, session)
            else:
                action = self._model.get_best_action(session, agent_observation_current)
        # Return the exploration action
        return action

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current):
        # Act with the best policy according to the model
        action = self._model.get_best_action(session, agent_observation_current)
        # Return the predicted action
        return action

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current,
                             agent_action, reward: float,
                             agent_observation_next,
                             warmup_step_current: int,
                             warmup_episode_current: int,
                             warmup_steps_volley: int):
        # Adjust the next observation if None (final step)
        last_step: bool = False
        if agent_observation_next is None:
            last_step = True
            if self._observation_space_type == SpaceType.discrete:
                agent_observation_next = 0
            else:
                agent_observation_next = numpy.zeros(self._observation_space_shape, dtype=float)
        # Choose randomly the next action
        self._next_warmup_action = interface.get_random_agent_action(logger, session)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current, agent_action, reward, agent_observation_next, self._next_warmup_action, last_step)

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current,
                            agent_action,
                            reward: float,
                            agent_observation_next,
                            train_step_current: int, train_step_absolute: int,
                            train_episode_current: int, train_episode_absolute: int,
                            train_episode_volley: int, train_episode_total: int):
        # Adjust the next observation if None (final step)
        last_step: bool = False
        if agent_observation_next is None:
            last_step = True
            if self._observation_space_type == SpaceType.discrete:
                agent_observation_next = 0
            else:
                agent_observation_next = numpy.zeros(self._observation_space_shape, dtype=float)
        # Choose an action according to the epsilon greedy approach: best action predicted by the model or random action
        if self._exploration_rate > 0 and random.uniform(0, 1.0) < self._exploration_rate:
            self._next_train_action = interface.get_random_agent_action(logger, session)
        else:
            self._next_train_action = self._model.get_best_action(session, agent_observation_current)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current, agent_action, reward, agent_observation_next, self._next_train_action, last_step)
        # Update the model and save current loss and absolute errors
        summary, self._current_loss, self._current_absolute_errors = self._model.update(session, self._model.buffer.get(self._batch_size))
        # Update the buffer with the computed absolute error
        self._model.buffer.update(self._current_absolute_errors)
        # Update the summary at the absolute current step
        self._summary_writer.add_summary(summary, train_step_absolute)

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current,
                                agent_action,
                                reward: float,
                                agent_observation_next,
                                inference_step_current: int,
                                inference_episode_current: int,
                                inference_episode_volley: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: float,
                                episode_total_reward: float,
                                warmup_episode_current: int,
                                warmup_steps_volley: int):
        # Reset predicted next warmup action
        self._next_warmup_action = None

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: float,
                               episode_total_reward: float,
                               train_step_absolute: int,
                               train_episode_current: int, train_episode_absolute: int,
                               train_episode_volley: int, train_episode_total: int):
        # Reset predicted next train action
        self._next_train_action = None
        # Decrease the exploration rate by its decay value
        self._exploration_rate = max(self._exploration_rate_min, self._exploration_rate - self._exploration_rate_decay)

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: float,
                                   episode_total_reward: float,
                                   inference_episode_current: int,
                                   inference_episode_volley: int):
        pass

    @property
    def trainable_variables(self):
        # Return the trainable variables of the agent model in experiment/agent _scope
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        # Return the amount of warmup episodes required by the model
        return self._model.warmup_steps



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import tensorflow
import logging
import numpy

# Import required src

from usienarl import SpaceType


class Model:
    """
    Base abstract model class.
    A model is whatever is moving an agent, deciding actions according to one or more policies and being updated to
    better learn those policies.

    Before use, a model needs to be generated. This will make sure to define the internal structure of the model
    (usually a tensorflow graph).

    To define your own model, implement the abstract class in a specific child class.
    """

    def __init__(self,
                 name: str):
        # Define model attributes
        self._name: str = name
        # Define empty model attributes
        self._initializer = None
        self._scope: str = None
        self._observation_space_type: SpaceType = None
        self._observation_space_shape = None
        self._agent_action_space_type: SpaceType = None
        self._agent_action_space_shape = None
        self._supported_observation_space_types: [] = []
        self._supported_action_space_types: [] = []

    def generate(self,
                 logger: logging.Logger,
                 scope: str,
                 observation_space_type: SpaceType, observation_space_shape,
                 agent_action_space_type: SpaceType, agent_action_space_shape) -> bool:
        """
        Generate the tensorflow model with the scope given by the format: experiment_name/agent_name/model_name
        It calls the define (which is implemented on the child class) and define summary methods.

        This method is called every time the model is used in an agent on a new experiment or a new experiment iteration,
        since it makes sure that the model is rebuilt according to the experiment/agent scope, agent observations and
        actions spaces.

        :param logger: the logger used to print the model information, warnings and errors
        :param scope: the str _name of experiment/agent to use as a scope for the graph
        :param observation_space_type: the type of the agent observation space: discrete or continuous
        :param observation_space_shape: the shape of the agent observation space (it's a size if mono-dimensional)
        :param agent_action_space_type: the type of the agent action space: discrete or continuous
        :param agent_action_space_shape: the shape of the agent action space (it's a size if mono-dimensional)
        :return: True if the model generation is successful, False otherwise
        """
        logger.info("Generating model " + self._name + " with scope " + scope + "...")
        # Set model attributes
        self._observation_space_type = observation_space_type
        self._observation_space_shape = observation_space_shape
        self._agent_action_space_type = agent_action_space_type
        self._agent_action_space_shape = agent_action_space_shape
        self._scope = scope
        # Check whether or not the observation space and the action space types are supported by the model
        observation_space_type_supported: bool = False
        for space_type in self._supported_observation_space_types:
            if self._observation_space_type == space_type:
                observation_space_type_supported = True
                break
        if not observation_space_type_supported:
            logger.error("Error during generation of model: observation space type not supported")
            return False
        action_space_type_supported: bool = False
        for space_type in self._supported_action_space_types:
            if self._agent_action_space_type == space_type:
                action_space_type_supported = True
                break
        if not action_space_type_supported:
            logger.error("Error during generation of model: action space type not supported")
            return False
        logger.info("Model generation successful")
        # Define the tensorflow model graph
        self._define_graph()
        # Returns True to state success
        return True

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Initialize the variables of the model given the session.

        :param logger: the logger used to print the model information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Initialize the model running the session on the appropriate tensorflow operation
        session.run(self._initializer)
        logger.info("Model initialized to default state")

    @property
    def trainable_variables(self):
        """
        Get the trainable variables in the model (useful for saving the model or comparing the weights) given the
        current experiment/agent scope.

        :return: the trainable tensorflow variables.
        """
        # Get the training variables of the model under its scope: usually, the training variables of the tensorflow graph
        return tensorflow.trainable_variables(self._scope + "/" + self._name + "/")

    @property
    def name(self) -> str:
        """
        The name of the model.

        :return: the string name of the model.

        :return:
        """
        return self._name

    @property
    def scope(self) -> str:
        """
        The scope of the model.

        :return: the string scope of the model.
        """
        return self._scope

    def _define_graph(self):
        """
        Define the tensorflow graph of the model.

        It uses as _scope the format: experiment_name/agent_name/model_name
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    def update(self,
               session,
               batch: []):
        """
        Update the model weights (thus training the model) given a batch of samples, relative weights and a set of
        additional experiment parameters.

        :param session: the session of tensorflow currently running
        :param batch: a batch of samples each one consisting of a tuple at least comprising observation current, action and reward all wrapped in numpy arrays
        :return: the summary and a set of parameters (losses, errors, etc) depending on the model
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def warmup_steps(self) -> int:
        """
        Get the number of steps required to warm-up the model.

        :return: the integer number of warm-up steps required by the model
        """
        raise NotImplementedError()



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import tensorflow
import numpy
import logging
import matplotlib.pyplot as plot

# Import required src

from usienarl import Environment, Agent, Interface
from usienarl.interfaces import PassThroughInterface


class Experiment:
    """
    Base experiment abstract class.
    An experiment is a set of commands which allows to train or test or even both a certain agent on a certain
    environment using a certain interface to translate between the two.

    If an interface is not supplied, a default simple pass-through interface is used.

    To conduct an experiment, first make sure the setup phase is passed. Once an experiment is setup, the agent and
    the environment can be used.
    A default experiment consists of:
        - warm-up phase, where the agent acts in warm-up mode and which can be used or not depending on the agent
        inner model
        - volley phase, which consists in multiple sets of episodes dedicated to training after which the agent
        performance is validated
        - test phase, which consists in multiple cycles of test where the agent acts in inference mode and in which
        the agent performance is tested and which occurs only after validation is passed

    Each volley consist of two additional sub-phases:
        - train phase: where the agent acts in train mode and its internal model is updated
        - validation phase: where the agent acts in inference mode and its performance is tested

    To define your own experiment, implement the abstract class in a specific child class. A child class only requires
    to have implemented the methods used to validate and test the agent's model and conclude the experiment volley
    phase or the experiment as a whole.

    Two types of scores are produced, using the rewards given only by the environment:
        - total reward: the sum of all the rewards at each step in the episode, also called total episode reward
        - scaled reward: the average of all the rewards at each step in the episode, also called average step reward
    """

    def __init__(self,
                 name: str,
                 environment: Environment,
                 agent: Agent,
                 interface: Interface = None):
        # Define experiment attributes
        self.name: str = name
        self._environment: Environment = environment
        self._agent: Agent = agent
        self._interface: Interface = interface if interface is not None else PassThroughInterface(self._environment)
        # Define empty experiment attributes
        self._agent_saver = None
        self._checkpoint_path: str = None
        self._metagraph_path: str = None
        self._plots_path: str = None
        self._tensorflow_config = None
        self._current_id: str = None
        self._trained_steps: int = None
        self._trained_episodes: int = None
        self._scaled_rewards: [] = []

    def setup(self,
              summary_path: str, metagraph_path: str,
              logger: logging.Logger,
              checkpoint_path: str = None, scopes_to_restore: [] = None,
              iteration: int = -1) -> bool:
        """
        Setup the experiment, preparing all of its component to execution. This must be called before conduct method.

        If an already trained model checkpoint is given, the last checkpoint in the given path folder is used to further
        train the model. The scope of the experiment and of the agent has to be the same across both models or a list
        of scopes to restore has to be defined. Using that list, which is passed to the agent, it is possible to decide
        which variables to restore using the appropriate agent's property.

        :param summary_path: the string path of the tensorboard summary directory to save during model training, set to None if the model is not being trained
        :param metagraph_path: the string path of the metagraph agent directory to save at the end of each train volley, set to None if the model is not being trained
        :param logger: the logger used to print the experiment information, warnings and errors
        :param checkpoint_path: the optional checkpoint path to load the pre-trained model from, it should be a valid model, default to None
        :param scopes_to_restore: the scopes of the checkpoint path to restore into the model, it can be used to define what part of the model should be restored
        :param iteration: number to append to the experiment name in all scopes and print statements (if not less than zero)
        :return: a boolean equals to True if the setup of the experiment is successful, False otherwise
        """
        # Save current experiment id using name and iteration
        self._current_id = self.name + "_" + str(iteration) if iteration > -1 else self.name
        # Start setup process
        logger.info("Setup of experiment " + self._current_id + "...")
        # Reset the tensorflow graph to a blank new graph
        tensorflow.reset_default_graph()
        # Execute the setup on the environment and on the agent and return if not both of them setup correctly
        logger.info("Environment setup...")
        if not self._environment.setup(logger):
            logger.info("Environment setup failed. Cannot setup the experiment!")
            return False
        logger.info("Environment setup is successful!")
        logger.info("Agent setup...")
        if not self._agent.setup(logger,
                                 self._interface.observation_space_type, self._interface.observation_space_shape,
                                 self._interface.agent_action_space_type, self._interface.agent_action_space_shape,
                                 self._current_id, summary_path, checkpoint_path, scopes_to_restore):
            logger.info("Agent setup failed. Cannot setup the experiment!")
            return False
        logger.info("Agent setup is successful!")
        # If setup of both environment and agent is successful, reset internal experiment variables
        self._metagraph_path = metagraph_path
        self._plots_path = self._metagraph_path.replace('/metagraph', '/plots')
        self._trained_steps: int = 0
        self._trained_episodes: int = 0
        # Initialize the agent internal model saver
        self._checkpoint_path = checkpoint_path
        self._agent_saver = tensorflow.train.Saver(self._agent.trainable_variables)
        if self._metagraph_path is not None:
            logger.info("Agent internal model will be saved after each train volley")
            logger.info("Agent internal model metagraph save path: " + self._metagraph_path)
        # Initialize tensorflow gpu configuration
        self._tensorflow_config = tensorflow.ConfigProto()
        self._tensorflow_config.gpu_options.allow_growth = True
        logger.info("Tensorflow configuration is successful!")
        # Setup is successful
        logger.info("Experiment setup is successful. Now can conduct the experiment")
        return True

    def _warmup(self,
                logger: logging.Logger,
                steps: int, episode_length: int,
                session,
                render: bool = False):
        """
        Execute a volley of warm-up of the agent on the environment. This volley is measured in steps and it is usually
        run for how many steps are required by the model.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param steps: the number of steps for which to run the agent in warm-up mode
        :param episode_length: the max length in steps of each episode
        :param session: the session of tensorflow currently running
        :param render: boolean parameter deciding whether or not to render during warm-up
        """
        episode: int = 0
        warmup_step: int = 0
        while True:
            # Start the episode
            episode = episode + 1
            episode_rewards: [] = []
            state_current = self._environment.reset(logger, session)
            # Execute actions until the episode is completed or the maximum length is exceeded
            for step in range(episode_length):
                # Increase counter and check if we can stop warming up
                warmup_step = warmup_step + 1
                if warmup_step > steps:
                    return
                # Print current warming-up progress every once in a while (if length is not too short)
                if steps > 100 and warmup_step % (steps // 10) == 0 and warmup_step > 0:
                    logger.info("Warmed-up for " + str(warmup_step) + "/" + str(steps) + " steps...")
                # Get the action decided by the agent with train policy
                observation_current = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action = self._agent.act_warmup(logger, session, self._interface, observation_current)
                # Get the next state with relative reward and completion flag
                environment_action = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, environment_action, session)
                # Complete the step and send back information to the agent
                observation_next = self._interface.environment_state_to_observation(logger, session, state_next)
                # Set the observation next to None if final state is reached
                if episode_done:
                    observation_next = None
                self._agent.complete_step_warmup(logger, session, self._interface,
                                                 observation_current, agent_action, reward, observation_next,
                                                 step, episode, steps)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Add the reward to the list of rewards for this episode
                episode_rewards.append(reward)
                # Update the current state with the previously next state
                state_current = state_next
                # Check if the episode is completed
                if episode_done:
                    break
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_warmup(logger, session, self._interface,
                                                episode_rewards[-1], numpy.sum(numpy.array(episode_rewards)),
                                                episode, steps)

    def _train(self,
               logger: logging.Logger,
               episodes: int, episode_length: int, episodes_max: int,
               session,
               render: bool = False):
        """
        Execute a volley of training of the agent on the environment.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param episodes: the number of episodes in which to run the agent in train mode
        :param episode_length: the max length in steps of each episode
        :param episodes_max: the max number of allowed training episodes
        :param session: the session of tensorflow currently running
        :param render: boolean parameter deciding whether or not to render during training
        :return: the list of rewards obtained and episode steps performed in the volley, grouped by episode
        """
        # Define a list of rewards/steps found in the volley
        volley_rewards: [] = []
        volley_steps: [] = []
        for episode in range(episodes):
            # Print current training progress every once in a while (if length is not too short)
            if episodes > 100 and episode % (episodes // 10) == 0 and episode > 0:
                logger.info("Trained for " + str(episode) + "/" + str(episodes) + " episodes...")
            # Start the episode
            episode_rewards: [] = []
            state_current = self._environment.reset(logger, session)
            # Execute actions until the episode is completed or the maximum length is exceeded
            for step in range(episode_length):
                # Get the action decided by the agent with train policy
                observation_current = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action = self._agent.act_train(logger, session, self._interface, observation_current)
                # Get the next state with relative reward and completion flag
                environment_action = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, environment_action, session)
                # Complete the step and send back information to the agent
                observation_next = self._interface.environment_state_to_observation(logger, session, state_next)
                # Set the observation next to None if final state is reached
                if episode_done:
                    observation_next = None
                self._agent.complete_step_train(logger, session, self._interface,
                                                observation_current, agent_action, reward, observation_next,
                                                step, self._trained_steps,
                                                episode, self._trained_episodes,
                                                episodes, episodes_max)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Add the reward to the list of rewards for this episode
                episode_rewards.append(reward)
                # Update the current state with the previously next state
                state_current = state_next
                # Increase the number of trained steps
                self._trained_steps += 1
                # Save volley steps up to termination point
                if step + 1 == episode_length or episode_done:
                    volley_steps.append(step)
                # Check if the episode is completed
                if episode_done:
                    break
            # Add the list of episode rewards to the volley rewards
            volley_rewards.append(episode_rewards)
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_train(logger, session, self._interface,
                                               episode_rewards[-1], numpy.sum(numpy.array(episode_rewards)),
                                               self._trained_steps,
                                               episode, self._trained_episodes,
                                               episodes, episodes_max)
            # Increase the counter of trained episodes
            self._trained_episodes += 1
        # Return the list of all rewards/steps seen in the volley
        return volley_rewards, volley_steps

    def _inference(self,
                   logger: logging.Logger,
                   episodes: int, episode_length: int,
                   session,
                   render: bool = False) -> []:
        """
        Execute a volley of inference of the agent on the environment.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param episodes: the number of episodes in which to run the agent in inference mode
        :param episode_length: the max length in steps of each episode
        :param session: the session of tensorflow currently running
        :param render: boolean parameter deciding whether or not to render during inference
        :return: the list of rewards obtained and episode step performed in the volley, grouped by episode
        """
        # Define a list of rewards/steps found in the volley
        volley_rewards: [] = []
        volley_steps: [] = []
        for episode in range(episodes):
            # Print current validation progress every once in a while (if length is not too short)
            if episodes > 100 and episode % (episodes // 10) == 0 and episode > 0:
                logger.info("Validated for " + str(episode) + "/" + str(episodes) + " episodes...")
            # Start the episode
            episode_rewards: [] = []
            state_current = self._environment.reset(logger, session)
            # Execute actions until the episode is completed or the maximum length is exceeded
            for step in range(episode_length):
                # Get the action decided by the agent with train policy
                observation_current = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action = self._agent.act_inference(logger, session, self._interface, observation_current)
                # Get the next state with relative reward and completion flag
                environment_action = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, environment_action, session)
                # Complete the step and send back information to the agent
                observation_next = self._interface.environment_state_to_observation(logger, session, state_next)
                # Set the observation next to None if final state is reached
                if episode_done:
                    observation_next = None
                self._agent.complete_step_inference(logger, session, self._interface,
                                                    observation_current, agent_action, reward, observation_next,
                                                    step, episode, episodes)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Add the reward to the list of rewards for this episode
                episode_rewards.append(reward)
                # Update the current state with the previously next state
                state_current = state_next
                # Save volley steps up to termination point
                if step + 1 == episode_length or episode_done:
                    volley_steps.append(step)
                # Check if the episode is completed
                if episode_done:
                    break
            # Add the list of episode rewards to the volley rewards
            volley_rewards.append(episode_rewards)
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_inference(logger, session, self._interface,
                                                   episode_rewards[-1], numpy.sum(numpy.array(episode_rewards)),
                                                   episode, episodes)
        # Return the list of all rewards/steps seen in the volley
        return volley_rewards, volley_steps

    def conduct(self,
                training_episodes_per_volley: int, validation_episodes_per_volley: int,
                training_episodes_max: int, episode_length_max: int,
                test_episodes_per_cycle: int, test_cycles: int,
                logger: logging.Logger,
                render_during_training: bool = False, render_during_validation: bool = False, render_during_test: bool = False,
                plot_sample_density_training: int = 1, plot_sample_density_validation: int = 1):
        """
        Conduct the experiment of the given number of training (up to the given maximum) and validation episodes
        per volley, with the associated given number of test episodes per cycle and the given number of test cycles after
        the validation has passed.

        A volley of training is always considered to stop the training exactly when the agent performs well enough
        to validate, by checking the appropriate condition. When validation is successful, appropriate condition is also
        checked on the test cycles to decide if the experiment is successful or not.

        Conducting an experiment will generate a tensorflow session for that experiment. It is required that the experiment
        is ready to be conducted by checking the result of the setup method before attempting at conducting it.

        :param training_episodes_per_volley: the number of training episodes per volley before trying to validate the agent
        :param validation_episodes_per_volley: the number of validation episodes per volley after the training in such interval
        :param training_episodes_max: the maximum number of training episodes allowed at all
        :param episode_length_max: the max length in steps of each episode
        :param test_episodes_per_cycle: the number of episodes to play for each test cycle after validation has passed
        :param test_cycles: the number of test cycles to execute after validation has passed
        :param logger: the logger used to print the experiment information, warnings and errors
        :param render_during_training: boolean flag to render the environment during training, default to False
        :param render_during_validation: boolean flag to render the environment during validation, default to False
        :param render_during_test: boolean flag to render the environment during test, default to False
        :param plot_sample_density_training: the optional number represent after how many episodes a sample for the episode-related training plots is sampled (default to 1, i.e. each episode)
        :param plot_sample_density_validation: the optional number represent after how many episodes a sample for the episode-related validation plots is sampled (default to 1, i.e. each episode)
        :return: the average and the max of the total average reward, the average and the max of the average average reward with respect to all test cycles, the training episodes required to validate, a success flag and the location in which the saved metagraph is stored
        """
        # Make the matplotlib use agg-backend to make it compatible with a server and increase the chunksize to avoid unwanted errors
        plot.switch_backend('agg')
        plot.rcParams['agg.path.chunksize'] = 1000
        # Start experiment
        logger.info("Conducting experiment " + self._current_id + "...")
        # Define the list in which to store the evolution of total and scaled average reward to plot by default
        all_training_episodes_total_rewards: [] = []
        all_training_episodes_scaled_rewards: [] = []
        all_training_episodes_length: [] = []
        all_validation_episodes_total_rewards: [] = []
        all_validation_episodes_scaled_rewards: [] = []
        all_validation_episodes_length: [] = []
        all_training_volleys_total_rewards: [] = []
        all_training_volleys_scaled_rewards: [] = []
        all_training_volleys_std_total_rewards: [] = []
        all_training_volleys_std_scaled_rewards: [] = []
        all_training_volleys_average_episode_length: [] = []
        all_validation_volleys_total_rewards: [] = []
        all_validation_volleys_scaled_rewards: [] = []
        all_validation_volleys_std_total_rewards: [] = []
        all_validation_volleys_std_scaled_rewards: [] = []
        all_validation_volleys_average_episode_length: [] = []
        # Define a counter for training volleys and validation volleys
        training_volley_counter: int = 0
        validation_volley_counter: int = 0
        # Define the session
        with tensorflow.Session(config=self._tensorflow_config) as session:
            # Initialize the environment and the agent
            self._environment.initialize(logger, session)
            self._agent.initialize(logger, session)
            self._environment.post_initialize(logger, session)
            # Load the pre-trained model if checkpoint path is given in setup
            if self._checkpoint_path is not None:
                checkpoint = tensorflow.train.get_checkpoint_state(self._checkpoint_path)
                # If checkpoint exists restore from checkpoint
                if checkpoint and checkpoint.model_checkpoint_path:
                    self._agent_saver.restore(session, tensorflow.train.latest_checkpoint(self._checkpoint_path))
                    logger.info("Model graph stored at " + self._checkpoint_path + " loaded successfully!")
                else:
                    logger.error("Checkpoint path specified is wrong: no model can be accessed at " + self._checkpoint_path)
            # Execute pre-training if the agent requires pre-training
            if self._agent.warmup_steps > 0:
                logger.info("Warming-up for " + str(self._agent.warmup_steps) + " steps...")
                self._warmup(logger,
                             self._agent.warmup_steps,
                             episode_length_max,
                             session)
            # Save last volley average training and validation total and scaled rewards
            last_training_volley_average_total_reward: float = 0.0
            last_training_volley_average_scaled_reward: float = 0.0
            last_training_volley_std_total_reward: float = 0.0
            last_training_volley_std_scaled_reward: float = 0.0
            last_validation_volley_average_total_reward: float = 0.0
            last_validation_volley_average_scaled_reward: float = 0.0
            last_validation_volley_std_total_reward: float = 0.0
            last_validation_volley_std_scaled_reward: float = 0.0
            # Save last volley training and validation rewards list
            last_training_volley_rewards: [] = []
            last_validation_volley_rewards: [] = []
            last_training_volley_steps: [] = []
            last_validation_volley_steps: [] = []
            # Execute training until max training episodes number is reached or the validation score is above the threshold
            while self._trained_episodes < training_episodes_max:
                # Run train for training episodes per volley and get the average score
                logger.info("Training for " + str(training_episodes_per_volley) + " episodes...")
                last_training_volley_rewards, last_training_volley_steps = self._train(logger,
                                                                                       training_episodes_per_volley,
                                                                                       episode_length_max,
                                                                                       training_episodes_max,
                                                                                       session,
                                                                                       render_during_training)
                training_volley_counter += 1
                # Since rewards list are of different sizes, get their sum and average
                training_volley_total_rewards: numpy.ndarray = numpy.zeros(len(last_training_volley_rewards), dtype=float)
                training_volley_scaled_rewards: numpy.ndarray = numpy.zeros(len(last_training_volley_rewards), dtype=float)
                for index, episode_rewards in enumerate(last_training_volley_rewards):
                    training_volley_total_rewards[index] = numpy.sum(numpy.array(episode_rewards))
                    training_volley_scaled_rewards[index] = numpy.average(numpy.array(episode_rewards))
                    # Update the global lists of all the average total and scaled rewards for each training episode
                    all_training_episodes_total_rewards.append(numpy.sum(numpy.array(episode_rewards)))
                    all_training_episodes_scaled_rewards.append(numpy.average(numpy.array(episode_rewards)))
                # Push all the saved steps into the episode list
                all_training_episodes_length += last_training_volley_steps
                # Compute average/std total reward and average/std scaled reward and episode length (in steps) over training volley and append it to the volley list
                last_training_volley_average_total_reward: float = numpy.round(numpy.average(training_volley_total_rewards), 3)
                last_training_volley_average_scaled_reward: float = numpy.round(numpy.average(training_volley_scaled_rewards), 3)
                last_training_volley_std_total_reward: float = numpy.round(numpy.std(training_volley_total_rewards), 3)
                last_training_volley_std_scaled_reward: float = numpy.round(numpy.std(training_volley_scaled_rewards), 3)
                last_training_volley_average_episode_length: int = numpy.rint(numpy.average(numpy.array(last_training_volley_steps)))
                all_training_volleys_total_rewards.append(last_training_volley_average_total_reward)
                all_training_volleys_scaled_rewards.append(last_training_volley_average_scaled_reward)
                all_training_volleys_std_total_rewards.append(last_training_volley_std_total_reward)
                all_training_volleys_std_scaled_rewards.append(last_training_volley_std_scaled_reward)
                all_training_volleys_average_episode_length.append(last_training_volley_average_episode_length)
                # Print training results
                logger.info("Training of " + str(training_episodes_per_volley) + " episodes finished with following result:")
                logger.info("Average total reward over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_training_volley_average_total_reward))
                logger.info("Standard deviation of total reward over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_training_volley_std_total_reward))
                logger.info("Average scaled reward over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_training_volley_average_scaled_reward))
                logger.info("Standard deviation of scaled reward over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_training_volley_std_scaled_reward))
                logger.info("Average episode length over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_training_volley_average_episode_length) + " steps")
                logger.info("Total environmental steps in training mode up to now: " + str(self._trained_steps))
                # Save the agent internal model at the current step
                logger.info("Saving the agent...")
                self._agent_saver.save(session, self._metagraph_path + "/" + self._current_id)
                # Run inference for validation episodes per volley and get the average score
                logger.info("Validating for " + str(validation_episodes_per_volley) + " episodes...")
                last_validation_volley_rewards, last_validation_volley_steps = self._inference(logger,
                                                                                               validation_episodes_per_volley,
                                                                                               episode_length_max,
                                                                                               session,
                                                                                               render_during_validation)
                validation_volley_counter += 1
                # Since rewards list are of different sizes, get their sum and average
                validation_volley_total_rewards: numpy.ndarray = numpy.zeros(len(last_validation_volley_rewards), dtype=float)
                validation_volley_scaled_rewards: numpy.ndarray = numpy.zeros(len(last_validation_volley_rewards), dtype=float)
                for index, episode_rewards in enumerate(last_validation_volley_rewards):
                    validation_volley_total_rewards[index] = numpy.sum(numpy.array(episode_rewards))
                    validation_volley_scaled_rewards[index] = numpy.average(numpy.array(episode_rewards))
                    # Update the global lists of all the average total and scale rewards for each validation episode
                    all_validation_episodes_total_rewards.append(numpy.sum(numpy.array(episode_rewards)))
                    all_validation_episodes_scaled_rewards.append(numpy.average(numpy.array(episode_rewards)))
                # Push all the saved steps into the episode list
                all_validation_episodes_length += last_validation_volley_steps
                # Compute average/std total reward and average/std scaled reward over validation volley and append it to volley list
                last_validation_volley_average_total_reward: float = numpy.round(numpy.average(validation_volley_total_rewards), 3)
                last_validation_volley_average_scaled_reward: float = numpy.round(numpy.average(validation_volley_scaled_rewards), 3)
                last_validation_volley_std_total_reward: float = numpy.round(numpy.std(validation_volley_total_rewards), 3)
                last_validation_volley_std_scaled_reward: float = numpy.round(numpy.std(validation_volley_scaled_rewards), 3)
                last_validation_volley_average_episode_length: int = numpy.rint(numpy.average(numpy.array(last_validation_volley_steps)))
                all_validation_volleys_total_rewards.append(last_validation_volley_average_total_reward)
                all_validation_volleys_scaled_rewards.append(last_validation_volley_average_scaled_reward)
                all_validation_volleys_std_total_rewards.append(last_validation_volley_std_total_reward)
                all_validation_volleys_std_scaled_rewards.append(last_validation_volley_std_scaled_reward)
                all_validation_volleys_average_episode_length.append(last_validation_volley_average_episode_length)
                # Print validation results
                logger.info("Validation of " + str(validation_episodes_per_volley) + " episodes finished with following result:")
                logger.info("Average total reward over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_validation_volley_average_total_reward))
                logger.info("Standard deviation of total reward over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_validation_volley_std_total_reward))
                logger.info("Average scaled reward over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_validation_volley_average_scaled_reward))
                logger.info("Standard deviation of scaled reward over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_validation_volley_std_scaled_reward))
                logger.info("Average episode length over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(last_validation_volley_average_episode_length) + " steps")
                # Check for validation
                if self._is_validated(logger,
                                      last_validation_volley_average_total_reward, last_validation_volley_average_scaled_reward,
                                      last_training_volley_average_total_reward, last_training_volley_average_scaled_reward,
                                      last_validation_volley_std_total_reward, last_validation_volley_std_scaled_reward,
                                      last_training_volley_std_total_reward, last_training_volley_std_scaled_reward,
                                      last_validation_volley_rewards, last_training_volley_rewards,
                                      plot_sample_density_training, plot_sample_density_validation):
                    logger.info("Validation of the agent is successful")
                    break
            logger.info("End of training")
            # Test the model and get all cycles total, scaled rewards and all rewards and all episode lengths
            test_average_total_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            test_average_scaled_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            test_std_total_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            test_std_scaled_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            test_average_episode_length: numpy.ndarray = numpy.zeros(test_cycles, dtype=int)
            test_cycles_rewards: [] = []
            for test in range(test_cycles):
                # Run inference for test episodes per cycles and get the average score
                logger.info("Cycle " + str(test + 1) + " - Testing for " + str(test_episodes_per_cycle) + " episodes...")
                test_cycle_rewards, test_cycle_steps = self._inference(logger,
                                                                       test_episodes_per_cycle,
                                                                       episode_length_max,
                                                                       session,
                                                                       render_during_test)
                # Append the test cycle rewards to the full list
                test_cycles_rewards.append(test_cycle_rewards)
                # Since rewards list are of different sizes, get their sum and average
                test_cycle_total_rewards: numpy.ndarray = numpy.zeros(len(test_cycle_rewards), dtype=float)
                test_cycle_scaled_rewards: numpy.ndarray = numpy.zeros(len(test_cycle_rewards), dtype=float)
                for index, episode_rewards in enumerate(test_cycle_rewards):
                    test_cycle_total_rewards[index] = numpy.sum(numpy.array(episode_rewards))
                    test_cycle_scaled_rewards[index] = numpy.average(numpy.array(episode_rewards))
                # Compute total and scaled reward over the test cycle
                test_cycle_average_total_reward: float = numpy.round(numpy.average(test_cycle_total_rewards), 3)
                test_cycle_average_scaled_reward: float = numpy.round(numpy.average(test_cycle_scaled_rewards), 3)
                test_cycle_std_total_reward: float = numpy.round(numpy.std(test_cycle_total_rewards), 3)
                test_cycle_std_scaled_reward: float = numpy.round(numpy.std(test_cycle_scaled_rewards), 3)
                # Compute average episode length over the test cycle
                test_cycle_average_episode_length: int = numpy.rint(numpy.average(numpy.array(test_cycle_steps)))
                # Display test additional optional statistics
                self._display_test_cycle_metrics(logger,
                                                 test_cycle_average_total_reward,
                                                 test_cycle_average_scaled_reward,
                                                 test_cycle_std_total_reward,
                                                 test_cycle_std_scaled_reward,
                                                 test_cycle_rewards,
                                                 plot_sample_density_training, plot_sample_density_validation)
                # Save the rewards and the episode average length
                test_average_total_rewards[test] = test_cycle_average_total_reward
                test_average_scaled_rewards[test] = test_cycle_average_scaled_reward
                test_std_total_rewards[test] = test_cycle_std_total_reward
                test_std_scaled_rewards[test] = test_cycle_std_scaled_reward
                test_average_episode_length[test] = test_cycle_average_episode_length
                # Print test results
                logger.info("Testing of " + str(test_episodes_per_cycle) + " episodes finished with following result:")
                logger.info("Average total reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_average_total_reward))
                logger.info("Standard deviation of total reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_std_total_reward))
                logger.info("Average scaled reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_average_scaled_reward))
                logger.info("Standard deviation of scaled reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_std_scaled_reward))
                logger.info("Average episode length over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_average_episode_length) + " steps")
            logger.info("End of test")
            # Get the average and the best total and scaled rewards over all cycles
            average_test_average_total_reward: float = numpy.round(numpy.average(test_average_total_rewards), 3)
            max_test_average_total_reward: float = numpy.round(numpy.max(test_average_total_rewards), 3)
            average_test_average_scaled_reward: float = numpy.round(numpy.average(test_average_scaled_rewards), 3)
            max_test_average_scaled_reward: float = numpy.round(numpy.max(test_average_scaled_rewards), 3)
            # Get the average and the minimum standard deviation of total and scaled rewards over all cycles
            average_test_std_total_reward: float = numpy.round(numpy.average(test_std_total_rewards), 3)
            min_test_std_total_reward: float = numpy.round(numpy.min(test_std_total_rewards), 3)
            average_test_std_scaled_reward: float = numpy.round(numpy.average(test_std_scaled_rewards), 3)
            min_test_std_scaled_reward: float = numpy.round(numpy.min(test_std_scaled_rewards), 3)
            # Get the average episode length over all cycles
            average_test_average_episode_length: int = numpy.rint(numpy.average(test_average_episode_length))
            # Print final results and outcome of the experiment
            logger.info("Average test total reward is " + str(average_test_average_total_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Average test scaled reward is " + str(average_test_average_scaled_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Max test total reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_average_total_reward))
            logger.info("Max test scaled reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_average_scaled_reward))
            logger.info("Average test standard deviation of total reward is " + str(average_test_std_total_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Average test standard deviation of scaled reward is " + str(average_test_std_scaled_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Min test standard deviation of total reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(min_test_std_total_reward))
            logger.info("Min test standard deviation of scaled reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(min_test_std_scaled_reward))
            logger.info("Average test episode length is " + str(average_test_average_episode_length) + " with " + str(self._trained_episodes) + " training episodes")
            # Save plots
            logger.info("Saving plots...")
            plot.plot(list(range(len(all_training_episodes_total_rewards)))[::plot_sample_density_training], all_training_episodes_total_rewards[::plot_sample_density_training], 'r-')
            plot.xlabel('Training episode')
            plot.ylabel('Total reward')
            plot.savefig(self._plots_path + "/training_episodes_total_rewards.png", dpi=300, transparent=True)
            plot.clf()
            plot.plot(list(range(len(all_training_episodes_scaled_rewards)))[::plot_sample_density_training], all_training_episodes_scaled_rewards[::plot_sample_density_training], 'r--')
            plot.xlabel('Training episode')
            plot.ylabel('Scaled reward')
            plot.savefig(self._plots_path + "/training_episodes_scaled_rewards.png", dpi=300, transparent=True)
            plot.clf()
            plot.plot(list(range(len(all_training_episodes_length)))[::plot_sample_density_training], all_training_episodes_length[::plot_sample_density_training], 'b-.')
            plot.xlabel('Training episode')
            plot.ylabel('Episode length (steps)')
            plot.savefig(self._plots_path + "/training_episodes_lengths.png", dpi=300, transparent=True)
            plot.clf()
            plot.plot(list(range(len(all_validation_episodes_total_rewards)))[::plot_sample_density_validation], all_validation_episodes_total_rewards[::plot_sample_density_validation], 'g-')
            plot.xlabel('Validation episode')
            plot.ylabel('Total reward')
            plot.savefig(self._plots_path + "/validation_episodes_total_rewards.png", dpi=300, transparent=True)
            plot.clf()
            plot.plot(list(range(len(all_validation_episodes_scaled_rewards)))[::plot_sample_density_validation], all_validation_episodes_scaled_rewards[::plot_sample_density_validation], 'g--')
            plot.xlabel('Validation episode')
            plot.ylabel('Scaled reward')
            plot.savefig(self._plots_path + "/validation_episodes_scaled_rewards.png", dpi=300, transparent=True)
            plot.clf()
            plot.plot(list(range(len(all_validation_episodes_length)))[::plot_sample_density_validation], all_validation_episodes_length[::plot_sample_density_validation], 'b-.')
            plot.xlabel('Validation episode')
            plot.ylabel('Episode length (steps)')
            plot.savefig(self._plots_path + "/validation_episodes_lengths.png", dpi=300, transparent=True)
            plot.clf()
            if training_volley_counter > 1:
                plot.plot(list(range(len(all_training_volleys_total_rewards))), all_training_volleys_total_rewards, 'r-')
                plot.xlabel('Training volley')
                plot.ylabel('Average total reward')
                plot.savefig(self._plots_path + "/training_volleys_total_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_training_volleys_scaled_rewards))), all_training_volleys_scaled_rewards, 'r--')
                plot.xlabel('Training volley')
                plot.ylabel('Average scaled reward')
                plot.savefig(self._plots_path + "/training_volleys_scaled_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_training_volleys_std_total_rewards))), all_training_volleys_std_total_rewards, 'c-')
                plot.xlabel('Training volley')
                plot.ylabel('Standard deviation of total reward')
                plot.savefig(self._plots_path + "/training_volleys_std_total_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_training_volleys_std_scaled_rewards))), all_training_volleys_std_scaled_rewards, 'c--')
                plot.xlabel('Training volley')
                plot.ylabel('Standard deviation of scaled reward')
                plot.savefig(self._plots_path + "/training_volleys_std_scaled_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_training_volleys_average_episode_length))), all_training_volleys_average_episode_length, 'b-.')
                plot.xlabel('Training volley')
                plot.ylabel('Average episode length (steps)')
                plot.savefig(self._plots_path + "/training_volleys_average_episode_length.png", dpi=300, transparent=True)
                plot.clf()
            if validation_volley_counter > 1:
                plot.plot(list(range(len(all_validation_volleys_total_rewards))), all_validation_volleys_total_rewards, 'm-')
                plot.xlabel('Validation volley')
                plot.ylabel('Average total reward')
                plot.savefig(self._plots_path + "/validation_volleys_total_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_validation_volleys_scaled_rewards))), all_validation_volleys_scaled_rewards, 'm--')
                plot.xlabel('Validation volley')
                plot.ylabel('Average scaled reward')
                plot.savefig(self._plots_path + "/validation_volleys_scaled_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_validation_volleys_std_total_rewards))), all_validation_volleys_std_total_rewards, 'y-')
                plot.xlabel('Validation volley')
                plot.ylabel('Standard deviation of total reward')
                plot.savefig(self._plots_path + "/validation_volleys_std_total_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_validation_volleys_std_scaled_rewards))), all_validation_volleys_std_scaled_rewards, 'y--')
                plot.xlabel('Validation volley')
                plot.ylabel('Standard deviation of scaled reward')
                plot.savefig(self._plots_path + "/validation_volleys_std_scaled_rewards.png", dpi=300, transparent=True)
                plot.clf()
                plot.plot(list(range(len(all_validation_volleys_average_episode_length))), all_validation_volleys_average_episode_length, 'b-.')
                plot.xlabel('Validation volley')
                plot.ylabel('Average episode length (steps)')
                plot.savefig(self._plots_path + "/validation_volleys_average_episode_length.png", dpi=300, transparent=True)
                plot.clf()
            logger.info("Plots saved successfully")
            # Check if the experiment is successful
            success: bool = self._is_successful(logger,
                                                average_test_average_total_reward, average_test_average_scaled_reward,
                                                max_test_average_total_reward, max_test_average_scaled_reward,
                                                average_test_std_total_reward, average_test_std_scaled_reward,
                                                min_test_std_total_reward, min_test_std_scaled_reward,
                                                last_validation_volley_average_total_reward,
                                                last_validation_volley_average_scaled_reward,
                                                last_training_volley_average_total_reward,
                                                last_training_volley_average_scaled_reward,
                                                last_validation_volley_std_total_reward,
                                                last_validation_volley_std_scaled_reward,
                                                last_training_volley_std_total_reward,
                                                last_training_volley_std_scaled_reward,
                                                test_cycles_rewards,
                                                last_validation_volley_rewards, last_training_volley_rewards,
                                                plot_sample_density_training, plot_sample_density_validation)
            if success:
                logger.info("The experiment terminated successfully")
            else:
                logger.info("The experiment terminated without meeting the requirements")
            return average_test_average_total_reward, max_test_average_total_reward, average_test_average_scaled_reward, max_test_average_scaled_reward, self._trained_episodes, success, self._metagraph_path

    def watch(self,
              episode_length_max: int,
              test_episodes_per_cycle: int, test_cycles: int,
              logger: logging.Logger,
              render: bool = True):
        """
        Watch an experiment with an already trained agent given by the checkpoint path defined during setup.
        Use this method to see how a saved agent model performs.

        :param episode_length_max: the max length in steps of each episode
        :param test_episodes_per_cycle: the number of episodes to play for each test cycle
        :param test_cycles: the number of test cycles to execute
        :param logger: the logger used to print the experiment information, warnings and errors
        :param render: boolean flag to render the environment, default to True
        """
        logger.info("Watching experiment " + self._current_id + "...")
        # Define the session
        with tensorflow.Session(config=self._tensorflow_config) as session:
            # Initialize the environment and the agent
            self._environment.initialize(logger, session)
            self._agent.initialize(logger, session)
            # Load the pre-trained model at the checkpoint path given in setup
            if self._checkpoint_path is None:
                logger.error("Need to specify a checkpoint path to watch the experiment!")
                return
            else:
                checkpoint = tensorflow.train.get_checkpoint_state(self._checkpoint_path)
                # If checkpoint exists restore from checkpoint
                if checkpoint and checkpoint.model_checkpoint_path:
                    self._agent_saver.restore(session, tensorflow.train.latest_checkpoint(self._checkpoint_path))
                    logger.info("Model graph stored at " + self._checkpoint_path + " loaded successfully!")
                else:
                    logger.error("Checkpoint path specified is wrong: no model can be accessed at " + self._checkpoint_path)
                    return
                # Test the model and get all cycles total, scaled rewards and all rewards and all episode lengths
                test_average_total_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
                test_average_scaled_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
                test_std_total_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
                test_std_scaled_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
                test_average_episode_length: numpy.ndarray = numpy.zeros(test_cycles, dtype=int)
                test_cycles_rewards: [] = []
                for test in range(test_cycles):
                    # Run inference for test episodes per cycles and get the average score
                    logger.info("Cycle " + str(test + 1) + " - Testing for " + str(test_episodes_per_cycle) + " episodes...")
                    test_cycle_rewards, test_cycle_steps = self._inference(logger,
                                                                           test_episodes_per_cycle,
                                                                           episode_length_max,
                                                                           session,
                                                                           render)
                    # Append the test cycle rewards to the full list
                    test_cycles_rewards.append(test_cycle_rewards)
                    # Since rewards list are of different sizes, get their sum and average
                    test_cycle_total_rewards: numpy.ndarray = numpy.zeros(len(test_cycle_rewards), dtype=float)
                    test_cycle_scaled_rewards: numpy.ndarray = numpy.zeros(len(test_cycle_rewards), dtype=float)
                    for index, episode_rewards in enumerate(test_cycle_rewards):
                        test_cycle_total_rewards[index] = numpy.sum(numpy.array(episode_rewards))
                        test_cycle_scaled_rewards[index] = numpy.average(numpy.array(episode_rewards))
                    # Compute total and scaled reward over the test cycle
                    test_cycle_average_total_reward: float = numpy.average(test_cycle_total_rewards)
                    test_cycle_average_scaled_reward: float = numpy.average(test_cycle_scaled_rewards)
                    test_cycle_std_total_reward: float = numpy.round(numpy.std(test_cycle_total_rewards), 3)
                    test_cycle_std_scaled_reward: float = numpy.round(numpy.std(test_cycle_scaled_rewards), 3)
                    # Compute average episode length over the test cycle
                    test_cycle_average_episode_length: int = numpy.rint(numpy.average(numpy.array(test_cycle_steps)))
                    # Display test additional optional statistics
                    self._display_test_cycle_metrics(logger,
                                                     test_cycle_average_total_reward,
                                                     test_cycle_average_scaled_reward,
                                                     test_cycle_std_total_reward,
                                                     test_cycle_std_scaled_reward,
                                                     test_cycle_rewards)
                    # Save the rewards
                    test_average_total_rewards[test] = test_cycle_average_total_reward
                    test_average_scaled_rewards[test] = test_cycle_average_scaled_reward
                    test_std_total_rewards[test] = test_cycle_std_total_reward
                    test_std_scaled_rewards[test] = test_cycle_std_scaled_reward
                    test_average_episode_length[test] = test_cycle_average_episode_length
                    # Print test results
                    logger.info("Testing of " + str(test_episodes_per_cycle) + " episodes finished with following result:")
                    logger.info("Average total reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_average_total_reward))
                    logger.info("Standard deviation of total reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_std_total_reward))
                    logger.info("Average scaled reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_average_scaled_reward))
                    logger.info("Standard deviation of scaled reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_std_scaled_reward))
                    logger.info("Average episode length over " + str(test_episodes_per_cycle) + " test episodes: " + str(test_cycle_average_episode_length) + " steps")
                # Get the average and the best total and scaled rewards over all cycles
                average_test_average_total_reward: float = numpy.round(numpy.average(test_average_total_rewards), 3)
                max_test_average_total_reward: float = numpy.round(numpy.max(test_average_total_rewards), 3)
                average_test_average_scaled_reward: float = numpy.round(numpy.average(test_average_scaled_rewards), 3)
                max_test_average_scaled_reward: float = numpy.round(numpy.max(test_average_scaled_rewards), 3)
                # Get the average and the minimum standard deviation of total and scaled rewards over all cycles
                average_test_std_total_reward: float = numpy.round(numpy.average(test_std_total_rewards), 3)
                min_test_std_total_reward: float = numpy.round(numpy.min(test_std_total_rewards), 3)
                average_test_std_scaled_reward: float = numpy.round(numpy.average(test_std_scaled_rewards), 3)
                min_test_std_scaled_reward: float = numpy.round(numpy.min(test_std_scaled_rewards), 3)
                # Get the average episode length over all cycles
                average_test_average_episode_length: int = numpy.rint(numpy.average(test_average_episode_length))
                # Print final results
                logger.info("Average test total reward is " + str(average_test_average_total_reward))
                logger.info("Average test scaled reward is " + str(average_test_average_scaled_reward))
                logger.info("Max test total reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_average_total_reward))
                logger.info("Max test scaled reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_average_scaled_reward))
                logger.info("Average test standard deviation of total reward is " + str(average_test_std_total_reward))
                logger.info("Average test standard deviation of scaled reward is " + str(average_test_std_scaled_reward))
                logger.info("Min test standard deviation of total reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(min_test_std_total_reward))
                logger.info("Min test standard deviation of scaled reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(min_test_std_scaled_reward))
                logger.info("Average test episode length is " + str(average_test_average_episode_length))

    def initialize(self):
        """
        Initialize the experiment, resetting all class-specific variables and preparing for the next iteration of the same
        experiment setup.
        """
        raise NotImplementedError()

    def _is_validated(self,
                      logger: logging.Logger,
                      last_average_validation_total_reward: float, last_average_validation_scaled_reward: float,
                      last_average_training_total_reward: float, last_average_training_scaled_reward: float,
                      last_std_validation_total_reward: float, last_std_validation_scaled_reward: float,
                      last_std_training_total_reward: float, last_std_training_scaled_reward: float,
                      last_validation_volley_rewards: [], last_training_volley_rewards: [],
                      plot_sample_density_training: int = 1, plot_sample_density_validation: int = 1) -> bool:
        """
        Check if the experiment is to be considered validated using the given parameters.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param last_average_validation_total_reward: the average total reward during last volley validation phase
        :param last_average_validation_scaled_reward: the average scaled reward during last volley validation phase
        :param last_average_training_total_reward: the average total reward during last volley training phase
        :param last_average_training_scaled_reward: the average scaled reward during last volley training phase
        :param last_std_validation_total_reward: the standard deviation of total reward during last volley validation phase
        :param last_std_validation_scaled_reward: the standard deviation of scaled reward during last volley validation phase
        :param last_std_training_total_reward: the standard deviation of total reward during last volley training phase
        :param last_std_training_scaled_reward: the standard deviation of scaled reward during last volley training phase
        :param last_validation_volley_rewards: the list of rewards (grouped by episode) for each step during last volley validation phase
        :param last_training_volley_rewards: the list of rewards (grouped by episode) for each step during last volley training phase
        :param plot_sample_density_training: the optional number represent after how many episodes a sample for the episode-related training plots is sampled (default to 1, i.e. each episode)
        :param plot_sample_density_validation: the optional number represent after how many episodes a sample for the episode-related validation plots is sampled (default to 1, i.e. each episode)
        :return: a boolean flag True if condition are satisfied, False otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    def _display_test_cycle_metrics(self,
                                    logger: logging.Logger,
                                    last_test_cycle_average_total_reward: float,
                                    last_test_cycle_average_scaled_reward: float,
                                    last_test_cycle_std_total_reward: float,
                                    last_test_cycle_std_scaled_reward: float,
                                    last_test_cycle_rewards: [],
                                    plot_sample_density_training: int = 1, plot_sample_density_validation: int = 1):
        """
        Display additional optional metrics from the last test cycle.
        Note: average and total and scaled reward over all the cycle test episodes are already displayed.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param last_test_cycle_average_total_reward: the average total reward in the last test cycle
        :param last_test_cycle_average_scaled_reward: the average scaled reward in the last test cycle
        :param last_test_cycle_std_total_reward: the standard deviation of total reward in the last test cycle
        :param last_test_cycle_std_scaled_reward: the standard deviation of scaled reward in the last test cycle
        :param last_test_cycle_rewards: a list of all the rewards obtained in each episode over all the episodes in the last test cycles
        :param plot_sample_density_training: the optional number represent after how many episodes a sample for the episode-related training plots is sampled (default to 1, i.e. each episode)
        :param plot_sample_density_validation: the optional number represent after how many episodes a sample for the episode-related validation plots is sampled (default to 1, i.e. each episode)
        """
        raise NotImplementedError()

    def _is_successful(self,
                       logger: logging.Logger,
                       average_test_total_reward: float, average_test_scaled_reward: float,
                       max_test_total_reward: float, max_test_scaled_reward: float,
                       average_test_std_total_reward: float, average_test_std_scaled_reward: float,
                       min_test_std_total_reward: float, min_test_std_scaled_reward: float,
                       last_average_validation_total_reward: float, last_average_validation_scaled_reward: float,
                       last_average_training_total_reward: float, last_average_training_scaled_reward: float,
                       last_std_validation_total_reward: float, last_std_validation_scaled_reward: float,
                       last_std_training_total_reward: float, last_std_training_scaled_reward: float,
                       test_cycles_rewards: [],
                       last_validation_volley_rewards: [], last_training_volley_rewards: [],
                       plot_sample_density_training: int = 1, plot_sample_density_validation: int = 1) -> bool:
        """
        Check if the experiment is to be considered successful using the given parameters.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param average_test_total_reward: the average total reward during all the test cycles
        :param average_test_scaled_reward: the average scaled reward during all the test cycles
        :param max_test_total_reward: the maximum total reward during all the test cycles
        :param max_test_scaled_reward: the maximum scaled reward during all the test cycles
        :param average_test_std_total_reward: the average standard deviation of total reward during all the test cycles
        :param average_test_std_scaled_reward: the average standard deviation of scaled reward during all the test cycles
        :param min_test_std_total_reward: the minimum standard deviation of total reward during all the test cycles
        :param min_test_std_scaled_reward: the minimum standard deviation of scaled reward during all the test cycles
        :param last_average_validation_total_reward: the average total reward during last volley validation phase
        :param last_average_validation_scaled_reward: the average scaled reward during last volley validation phase
        :param last_average_training_total_reward: the average total reward during last volley training phase
        :param last_average_training_scaled_reward: the average scaled reward during last volley training phase
        :param last_std_validation_total_reward: the standard deviation of total reward during last volley validation phase
        :param last_std_validation_scaled_reward: the standard deviation of scaled reward during last volley validation phase
        :param last_std_training_total_reward: the standard deviation of total reward during last volley training phase
        :param last_std_training_scaled_reward: the standard deviation of scaled reward during last volley training phase
        :param test_cycles_rewards: the list of rewards (grouped by episode) for each step during all the test cycles
        :param last_validation_volley_rewards: the list of rewards (grouped by episode) for each step during last volley validation phase
        :param last_training_volley_rewards: the list of rewards (grouped by episode) for each step during last volley training phase
        ::param plot_sample_density_training: the optional number represent after how many episodes a sample for the episode-related training plots is sampled (default to 1, i.e. each episode)
        :param plot_sample_density_validation: the optional number represent after how many episodes a sample for the episode-related validation plots is sampled (default to 1, i.e. each episode)
        :return: a boolean flag True if condition are satisfied, False otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import tensorflow

# Import required src

from usienarl import Interface, SpaceType


class Agent:
    """
    Base agent abstract class.

    An agent defines who or what operates in a certain environment during a certain experiment.
    An agent can act and being updated, but the inner working (model, policies and so on) are left to be decided by the
    implementation.

    An agent needs to be generated before running, and generation is done when executing agent setup.

    Any agent can act in three different modes:
        - warmup mode, before training (not all agents required that, it depends on the agent inner model)
        - train mode, during training
        - inference mode, when exploiting

    One or more policies can be defined for each one of these modes.

    To define your own agent, implement the abstract class in a specific child class.
    """

    def __init__(self,
                 name: str):
        # Define agent attributes
        self._name: str = name
        # Define empty agent attributes
        self._scope: str = None
        self._summary_writer = None
        self._checkpoint_path: str = None
        self._scopes_to_restore: [] = None
        self._observation_space_type: SpaceType = None
        self._observation_space_shape = None
        self._agent_action_space_type: SpaceType = None
        self._agent_action_space_shape = None

    def setup(self,
              logger: logging.Logger,
              observation_space_type: SpaceType, observation_space_shape,
              agent_action_space_type: SpaceType, agent_action_space_shape,
              scope: str, summary_path: str,
              checkpoint_path: str, scopes_to_restore: []) -> bool:
        """
        Setup the agent.
        It is called before the tensorflow session generation, if any.
        Note: this should generate the model and other components, if any.

        :param logger: the logger used to print the agent information, warnings and errors
        :param observation_space_type: the space type of the observation space
        :param observation_space_shape: the shape of the observation space
        :param agent_action_space_type: the space type of the agent action space
        :param agent_action_space_shape: the shape of the agent action space
        :param scope: the experiment scope encompassing the agent _scope, if any
        :param checkpoint_path: the path of the checkpoint to restore, if any
        :param scopes_to_restore: the list of scopes to restore with the checkpoint, if any
        :param summary_path: the path of the summary writer of the agent
        :return a boolean flag True if setup is successful, False otherwise
        """
        logger.info("Setup of agent " + self._name + " with scope " + scope + "...")
        # Reset agent attributes
        self._scope = scope
        self._checkpoint_path = checkpoint_path
        self._scopes_to_restore = scopes_to_restore
        self._observation_space_type: SpaceType = observation_space_type
        self._observation_space_shape = observation_space_shape
        self._agent_action_space_type: SpaceType = agent_action_space_type
        self._agent_action_space_shape = agent_action_space_shape
        # Reset the summary writer if required
        if summary_path is not None:
            self._summary_writer = tensorflow.summary.FileWriter(summary_path, graph=tensorflow.get_default_graph())
            logger.info("A Tensorboard summary for the agent will be updated during training of its internal model")
            logger.info("Tensorboard summary path: " + summary_path)
        # Try to generate the agent inner model
        return self._generate(logger, observation_space_type, observation_space_shape, agent_action_space_type, agent_action_space_shape)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape,
                  action_space_type: SpaceType, action_space_shape) -> bool:
        """
        Generate the agent internal model. Used to generate all custom components of the agent.
        It is always called during setup.

        :param logger: the logger used to print the agent information, warnings and errors
        :return a boolean flag True if setup is successful, False otherwise
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Initialize the agent before acting in the environment.
        It is called right after tensorflow session generation and after the environment is initialized.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current):
        """
        Take an action given the current agent observation in warmup mode. Usually it uses a random policy.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :return: the action the agent decided to take
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current):
        """
        Take an action given the current agent observation in train mode. Usually it uses an exploring policy.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :return: the action the agent decided to take
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current):
        """
        Take an action given the current agent observation in inference mode. Usually it uses the best possible policy.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :return: the action the agent decided to take
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current,
                             agent_action,
                             reward: float,
                             agent_observation_next,
                             warmup_step_current: int,
                             warmup_episode_current: int,
                             warmup_steps_volley: int):
        """
        Complete a warmup step with the given values.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :param agent_action: the action taken in the environment as seen by the agent leading from the current observation to the next observation
        :param reward: the reward obtained by the combination of current observation, next observation and action in-between
        :param agent_observation_next: the next observation of the agent
        :param warmup_step_current: the current warmup step in the current episode
        :param warmup_episode_current: the current warmup episode
        :param warmup_steps_volley: the number of warmup steps in the volley
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current,
                            agent_action,
                            reward: float,
                            agent_observation_next,
                            train_step_current: int, train_step_absolute: int,
                            train_episode_current: int, train_episode_absolute: int,
                            train_episode_volley: int, train_episode_total: int):
        """
        Complete a train step with the given values.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :param agent_action: the action taken in the environment as seen by the agent leading from the current observation to the next observation
        :param reward: the reward obtained by the combination of current observation, next observation and action in-between
        :param agent_observation_next: the next observation of the agent
        :param train_step_current: the current train step in the current episode
        :param train_step_absolute: the current absolute number of train step (counting all volleys)
        :param train_episode_current: the current train episode
        :param train_episode_absolute: the current absolute number of train episode (counting all volleys)
        :param train_episode_volley: the number of train episodes in the volley
        :param train_episode_total: the total number of allowed train episodes
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current,
                                agent_action,
                                reward: float,
                                agent_observation_next,
                                inference_step_current: int,
                                inference_episode_current: int,
                                inference_episode_volley: int):
        """
        Complete an inference step with the given values.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :param agent_action: the action taken in the environment as seen by the agent leading from the current observation to the next observation
        :param reward: the reward obtained by the combination of current observation, next observation and action in-between
        :param agent_observation_next: the next observation of the agent
        :param inference_step_current: the current inference step in the current episode
        :param inference_episode_current: the current inference episode
        :param inference_episode_volley: the number of inference episodes in the volley
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: float,
                                episode_total_reward: float,
                                warmup_episode_current: int,
                                warmup_steps_volley: int):
        """
        Finish a _warmup episode with the given values.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param last_step_reward: the reward obtained in the last step in the passed episode
        :param episode_total_reward: the reward obtained in the passed episode
        :param warmup_episode_current: the current warmup episode
        :param warmup_steps_volley: the number of warmup steps in the volley
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: float,
                               episode_total_reward: float,
                               train_step_absolute: int,
                               train_episode_current: int, train_episode_absolute: int,
                               train_episode_volley: int, train_episode_total: int):
        """
        Finish a train episode with the given values.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param last_step_reward: the reward obtained in the last step in the passed episode
        :param episode_total_reward: the reward obtained in the passed episode
        :param train_step_absolute: the current absolute number of train step (counting all volleys)
        :param train_episode_current: the current train episode
        :param train_episode_absolute: the current absolute number of train episode (counting all volleys)
        :param train_episode_volley: the number of train episodes in the volley
        :param train_episode_total: the total number of allowed train episodes
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: float,
                                   episode_total_reward: float,
                                   inference_episode_current: int,
                                   inference_episode_volley: int):
        """
        Finish an inference episode with the given values.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent and the environment
        :param last_step_reward: the reward obtained in the last step in the passed episode
        :param episode_total_reward: the reward obtained in the passed episode
        :param inference_episode_current: the current inference episode
        :param inference_episode_volley: the number of inference episodes in the volley
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def trainable_variables(self):
        """
        Get the trainable variables of the agent (usually of the internal model of the agent).
        Is is usually searched with the scope environment/agent.

        :return: the trainable variables defined by this agent.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def warmup_steps(self) -> int:
        """
        Return the integer number of warm-up steps required by the agent.

        :return: the integer number of warm-up steps required by the agent internal model
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()




#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.td_models import DuelingDeepQLearning
from usienarl.utils import softmax


class DuelingDeepQLearningAgentBoltzmann(Agent):
    """
    Dueling Double Deep Q-Learning agent with Boltzmann sampling exploration policy.
    It is supplied with a Dueling Double Deep Q-Learning model (DDDQN).

    The weight copy step interval defines after how many steps per interval the target network weights should be
    updated with the main network weights.

    The batch size define how many steps from the prioritized experience replay built-in buffer should be fed into
    the model when updating. The agent uses a 1-step temporal difference algorithm, i.e. it is updated at every step.

    During training, agent samples according to the value of the temperature. Such temperatures decays at every
    completed episode of its decay value, while remaining bounded by its defined max and min values.

    Note: default model is suited for simple environment without complex dynamics (for example, without an action mask).
    It is advised to implement a custom model for research task using default ones as templates.
    """

    def __init__(self,
                 name: str,
                 model: DuelingDeepQLearning,
                 weight_copy_step_interval: int,
                 batch_size: int = 1,
                 temperature_max: float = 1.0, temperature_min: float = 0.001,
                 temperature_decay: float = 0.001):
        # Define agent attributes
        self._model: DuelingDeepQLearning = model
        self._temperature_max: float = temperature_max
        self._temperature_min: float = temperature_min
        self._temperature_decay: float = temperature_decay
        # Define internal agent attributes
        self._weight_copy_step_interval: int = weight_copy_step_interval
        self._batch_size: int = batch_size
        self._current_absolute_errors = None
        self._current_loss = None
        self._temperature: float = None
        # Generate base agent
        super(DuelingDeepQLearningAgentBoltzmann, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape,
                  agent_action_space_type: SpaceType, agent_action_space_shape) -> bool:
        # Generate the model and return a flag stating if generation was successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset internal agent attributes
        self._current_absolute_errors = None
        self._current_loss = None
        # Initialize the model
        self._model.initialize(logger, session)
        # Reset temperature to its starting value (the max)
        self._temperature = self._temperature_max
        # Run the weight copy operation to uniform main and target networks
        self._model.copy_weight(session)

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current):
        # Act randomly
        action = interface.get_random_agent_action(logger, session)
        # Return the random action
        return action

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current):
        # Act according to boltzmann approach: get the softmax over all the actions predicted by the model
        output = softmax(self._model.get_all_action_values(session, agent_observation_current) / self._temperature)
        # Make sure output sums up to 1.0
        output = output / output.sum()
        # Get a random action value (random output) using the softmax as probability distribution
        action_value = numpy.random.choice(output[0], p=output[0])
        # Return the chosen action as the index of such chosen action value
        return numpy.argmax(output[0] == action_value)

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current):
        # Act with the best policy according to the model
        action = self._model.get_best_action(session, agent_observation_current)
        # Return the predicted action
        return action

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current,
                             agent_action, reward: float,
                             agent_observation_next,
                             warmup_step_current: int,
                             warmup_episode_current: int,
                             warmup_steps_volley: int):
        # Adjust the next observation if None (final step)
        last_step: bool = False
        if agent_observation_next is None:
            last_step = True
            if self._observation_space_type == SpaceType.discrete:
                agent_observation_next = 0
            else:
                agent_observation_next = numpy.zeros(self._observation_space_shape, dtype=float)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current, agent_action, reward, agent_observation_next, last_step)

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current,
                            agent_action,
                            reward: float,
                            agent_observation_next,
                            train_step_current: int, train_step_absolute: int,
                            train_episode_current: int, train_episode_absolute: int,
                            train_episode_volley: int, train_episode_total: int):
        # Adjust the next observation if None (final step)
        last_step: bool = False
        if agent_observation_next is None:
            last_step = True
            if self._observation_space_type == SpaceType.discrete:
                agent_observation_next = 0
            else:
                agent_observation_next = numpy.zeros(self._observation_space_shape, dtype=float)
        # After each weight step interval update the target network weights with the main network weights
        if (train_step_absolute % self._weight_copy_step_interval) == 0 and train_episode_absolute > 0:
            logger.info("Copying weights from main network to target network at step " + str(train_step_absolute))
            self._model.copy_weight(session)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current, agent_action, reward, agent_observation_next, last_step)
        # Update the model and save current loss and absolute errors
        summary, self._current_loss, self._current_absolute_errors = self._model.update(session, self._model.buffer.get(self._batch_size))
        # Update the buffer with the computed absolute error
        self._model.buffer.update(self._current_absolute_errors)
        # Update the summary at the absolute current step
        self._summary_writer.add_summary(summary, train_step_absolute)

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current,
                                agent_action,
                                reward: float,
                                agent_observation_next,
                                inference_step_current: int,
                                inference_episode_current: int,
                                inference_episode_volley: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: float,
                                episode_total_reward: float,
                                warmup_episode_current: int,
                                warmup_steps_volley: int):
        pass

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: float,
                               episode_total_reward: float,
                               train_step_absolute: int,
                               train_episode_current: int, train_episode_absolute: int,
                               train_episode_volley: int, train_episode_total: int):
        # Decrease temperature rate by its decay value
        self._temperature = max(self._temperature_min, self._temperature - self._temperature_decay)

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: float,
                                   episode_total_reward: float,
                                   inference_episode_current: int,
                                   inference_episode_volley: int):
        pass

    @property
    def trainable_variables(self):
        # Return the trainable variables of the agent model in experiment/agent _scope
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        # Return the amount of warmup episodes required by the model
        return self._model.warmup_steps
